# Copyright (c) 2016-2023 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from torch_geometric.data import HeteroData, Data
import torch

from gqlalchemy.transformations.translators.translator import Translator
from gqlalchemy.transformations.constants import EDGE_INDEX, PYG_ID, NUM_NODES, DEFAULT_NODE_LABEL, DEFAULT_EDGE_TYPE
from gqlalchemy.memgraph_constants import (
    MG_HOST,
    MG_PORT,
    MG_USERNAME,
    MG_PASSWORD,
    MG_ENCRYPTED,
    MG_CLIENT_NAME,
    MG_LAZY,
)


class PyGTranslator(Translator):
    def __init__(
        self,
        host: str = MG_HOST,
        port: int = MG_PORT,
        username: str = MG_USERNAME,
        password: str = MG_PASSWORD,
        encrypted: bool = MG_ENCRYPTED,
        client_name: str = MG_CLIENT_NAME,
        lazy: bool = MG_LAZY,
    ) -> None:
        super().__init__(host, port, username, password, encrypted, client_name, lazy)

    @classmethod
    def get_node_properties(cls, graph, node_label: str, node_id: int):
        """Extracts node properties from heterogeneous graph based on the node_label.
        Args:
            graph: A reference to the PyG graph.
            node_label: Node label
            node_id: Node_id
        """
        node_properties = {}
        for iter_node_label, iter_node_properties in graph.node_items():
            if iter_node_label != node_label:
                continue
            for property_name, property_values in iter_node_properties.items():
                if property_name == NUM_NODES:
                    continue
                node_properties[property_name] = property_values[node_id]
        return node_properties

    @classmethod
    def extract_node_edge_properties_from_homogeneous_graph(cls, graph):
        """Homogenous graph don't have node and etype properties so it is hard to extract node and edge attributes.
        Args:
            graph: Data = reference to the PyG graph.
        Returns:
            node and edge attributes as dictionaries
        """
        graph_data = graph.to_dict()
        graph_data.pop(EDGE_INDEX, None)
        node_properties = dict(filter(lambda pair: graph.is_node_attr(pair[0]), graph_data.items()))
        etype_properties = dict(filter(lambda pair: graph.is_edge_attr(pair[0]), graph_data.items()))
        return node_properties, etype_properties

    def to_cypher_queries(self, graph):
        """Produce cypher queries for data saved as part of thePyG graph. The method handles both homogeneous and heterogeneous graph.
        The method converts 1D as well as multidimensional features. If there are some isolated nodes inside the graph, they won't get transferred. Nodes and edges
         created in Memgraph DB will, for the consistency reasons, have property `pyg_id` set to the id they have as part of the PyG graph. Note that this method doesn't insert anything inside
         the database, it just creates cypher queries. To insert queries the following code can be used:
         >>> memgraph = Memgraph()
         pyg_graph = HeteroData(...)
         for query in PyGTranslator().to_cypher_queries(pyg_graph):
            memgraph.execute(query)

         Args:
            graph: A reference to the PyG graph.
        Returns:
            cypher queries.
        """
        queries = []
        if graph is None:
            return queries

        if isinstance(graph, HeteroData):
            for etype, etype_features in graph.edge_items():
                source_node_label, edge_type, dest_node_label = etype
                # Get edges
                edge_index = etype_features[EDGE_INDEX]
                etype_features.pop(EDGE_INDEX, None)
                src_nodes, dest_nodes = edge_index[0], edge_index[1]
                eid = 0
                for src_node_id, dest_node_id in zip(src_nodes, dest_nodes):
                    # Get src node properties
                    source_node_properties = PyGTranslator.get_node_properties(graph, source_node_label, src_node_id)
                    source_node_properties[PYG_ID] = int(src_node_id)
                    # Destination node properties
                    dest_node_properties = PyGTranslator.get_node_properties(graph, dest_node_label, dest_node_id)
                    dest_node_properties[PYG_ID] = int(dest_node_id)
                    # Get edge features
                    edge_properties = Translator.get_properties(etype_features, eid)
                    edge_properties[PYG_ID] = eid
                    eid += 1
                    # Create query
                    queries.append(
                        self.create_insert_query(
                            source_node_label,
                            source_node_properties,
                            edge_type,
                            edge_properties,
                            dest_node_label,
                            dest_node_properties,
                        )
                    )
        elif isinstance(graph, Data):
            # Get edges
            src_nodes, dest_nodes = graph.edge_index
            # Get graph data
            node_properties, etype_properties = PyGTranslator.extract_node_edge_properties_from_homogeneous_graph(graph)
            # Process edges
            eid = 0
            for src_node_id, dest_node_id in zip(src_nodes, dest_nodes):
                # Get node properties
                source_node_properties = Translator.get_properties(node_properties, src_node_id)
                source_node_properties[PYG_ID] = int(src_node_id)
                dest_node_properties = Translator.get_properties(node_properties, dest_node_id)
                dest_node_properties[PYG_ID] = int(dest_node_id)
                edge_properties = Translator.get_properties(etype_properties, eid)
                edge_properties[PYG_ID] = eid
                eid += 1
                queries.append(
                    self.create_insert_query(
                        DEFAULT_NODE_LABEL,
                        source_node_properties,
                        DEFAULT_EDGE_TYPE,
                        edge_properties,
                        DEFAULT_NODE_LABEL,
                        dest_node_properties,
                    )
                )

        return queries

    def get_instance(self):
        """Create instance of PyG graph from all edges that are inside Memgraph. Currently, isolated nodes are ignored because they don't contribute in message passing neural networks. Only numerical features
        that are set on all nodes or all edges are transferred to the PyG instance since this is PyG's requirement. That means that any string values properties won't be transferred, as well as numerical properties
        that aren't set on all nodes. However, features that are of type list are transferred to the PyG instance and can be used as any other feature in the PyG graph. Regardless of data residing inside Memgraph database, the created
        PyG graph is a heterograph instance.
        Returns:
            PyG heterograph instance.
        """
        # Parse into nice data structures
        src_nodes, dest_nodes, node_features, edge_features, mem_indexes = self._parse_memgraph()

        # Create PyG heterograph
        graph = HeteroData()

        # Create edges in COO format
        for type_triplet in src_nodes.keys():
            graph[type_triplet].edge_index = torch.tensor(
                [src_nodes[type_triplet], dest_nodes[type_triplet]], dtype=torch.int32
            )

        # Set number of nodes, otherwise PyG infers automatically and warnings can occur
        for node_label, num_nodes_label in mem_indexes.items():
            graph[node_label].num_nodes = num_nodes_label

        # Set node features
        for node_label, features_dict in node_features.items():
            for feature_name, features in features_dict.items():
                features = Translator.validate_features(features, graph[node_label].num_nodes)
                if features is not None:
                    setattr(graph[node_label], feature_name, torch.tensor(features, dtype=torch.float32))

        # Set edge features
        for edge_triplet, features_dict in edge_features.items():
            for feature_name, features in features_dict.items():
                features = Translator.validate_features(features, graph[edge_triplet].num_edges)
                if features is not None:
                    setattr(graph[edge_triplet], feature_name, torch.tensor(features, dtype=torch.float32))

        # PyG offers a method to validate graph so if something is wrong an error will occur
        graph.validate(raise_on_error=True)
        return graph



# Copyright (c) 2016-2022 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Any, Dict, Iterator, List, Optional

from gqlalchemy.connection import Connection
from gqlalchemy.exceptions import GQLAlchemyError
from gqlalchemy.models import (
    Constraint,
    Index,
    Node,
    Relationship,
)


class DatabaseClient(ABC):
    def __init__(
        self,
        host: str,
        port: int,
        username: str,
        password: str,
        encrypted: bool,
        client_name: str,
    ):
        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self._encrypted = encrypted
        self._client_name = client_name
        self._cached_connection: Optional[Connection] = None

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    def execute_and_fetch(
        self, query: str, parameters: Dict[str, Any] = {}, connection: Connection = None
    ) -> Iterator[Dict[str, Any]]:
        """Executes Cypher query and returns iterator of results."""
        connection = connection or self._get_cached_connection()
        return connection.execute_and_fetch(query, parameters)

    def execute(self, query: str, parameters: Dict[str, Any] = {}, connection: Connection = None) -> None:
        """Executes Cypher query without returning any results."""
        connection = connection or self._get_cached_connection()
        connection.execute(query, parameters)

    def create_index(self, index: Index) -> None:
        """Creates an index (label or label-property type) in the database."""
        query = f"CREATE INDEX ON {index.to_cypher()};"
        self.execute(query)

    def drop_index(self, index: Index) -> None:
        """Drops an index (label or label-property type) in the database."""
        query = f"DROP INDEX ON {index.to_cypher()};"
        self.execute(query)

    @abstractmethod
    def get_indexes(self) -> List[Index]:
        """Returns a list of all database indexes (label and label-property types)."""
        pass

    @abstractmethod
    def ensure_indexes(self, indexes: List[Index]) -> None:
        """Ensures that database indexes match input indexes."""
        pass

    def drop_indexes(self) -> None:
        """Drops all indexes in the database"""
        self.ensure_indexes(indexes=[])

    def create_constraint(self, index: Constraint) -> None:
        """Creates a constraint (label or label-property type) in the database."""
        query = f"CREATE CONSTRAINT ON {index.to_cypher()};"
        self.execute(query)

    def drop_constraint(self, index: Constraint) -> None:
        """Drops a constraint (label or label-property type) in the database."""
        query = f"DROP CONSTRAINT ON {index.to_cypher()};"
        self.execute(query)

    @abstractmethod
    def get_constraints(
        self,
    ) -> List[Constraint]:
        """Returns a list of all database constraints (label and label-property types)."""
        pass

    @abstractmethod
    def get_exists_constraints(
        self,
    ) -> List[Constraint]:
        pass

    @abstractmethod
    def get_unique_constraints(
        self,
    ) -> List[Constraint]:
        pass

    def ensure_constraints(
        self,
        constraints: List[Constraint],
    ) -> None:
        """Ensures that database constraints match input constraints."""
        old_constraints = set(self.get_constraints())
        new_constraints = set(constraints)
        for obsolete_constraints in old_constraints.difference(new_constraints):
            self.drop_constraint(obsolete_constraints)
        for missing_constraint in new_constraints.difference(old_constraints):
            self.create_constraint(missing_constraint)

    def drop_database(self):
        """Drops database by removing all nodes and edges."""
        self.execute("MATCH (n) DETACH DELETE n;")

    def _get_cached_connection(self) -> Connection:
        """Returns cached connection if it exists, creates it otherwise."""
        if self._cached_connection is None or not self._cached_connection.is_active():
            self._cached_connection = self.new_connection()

        return self._cached_connection

    @abstractmethod
    def new_connection(self) -> Connection:
        """Creates new database connection."""
        pass

    def _get_nodes_with_unique_fields(self, node: Node) -> Optional[Node]:
        """Get's all nodes from the database that have any of the unique fields
        set to the values in the `node` object.
        """
        return self.execute_and_fetch(
            f"MATCH (node: {node._label})" f" WHERE {node._get_cypher_unique_fields_or_block('node')}" f" RETURN node;"
        )

    def get_variable_assume_one(self, query_result: Iterator[Dict[str, Any]], variable_name: str) -> Any:
        """Returns a single result from the query_result (usually gotten from
        the execute_and_fetch function).
        If there is more than one result, raises a GQLAlchemyError.
        """
        result = next(query_result, None)
        next_result = next(query_result, None)
        if result is None:
            raise GQLAlchemyError("No result found. Result list is empty.")
        elif next_result is not None:
            raise GQLAlchemyError(
                f"One result expected, but more than one result found. First result: {result}, second result: {next_result}"
            )
        elif variable_name not in result:
            raise GQLAlchemyError(f"Variable name {variable_name} not present in result.")

        return result[variable_name]

    def create_node(self, node: Node) -> Optional[Node]:
        """Creates a node in database from the `node` object."""
        results = self.execute_and_fetch(
            f"CREATE (node:{node._label}) {node._get_cypher_set_properties('node')} RETURN node;"
        )
        return self.get_variable_assume_one(results, "node")

    @abstractmethod
    def save_node(self, node: Node) -> Node:
        """Saves node to database.
        If the node._id is not None, it fetches the node with the same id from
        the database and updates it's fields.
        If the node has unique fields it fetches the nodes with the same unique
        fields from the database and updates it's fields.
        Otherwise it creates a new node with the same properties.
        Null properties are ignored.
        """
        pass

    def save_nodes(self, nodes: List[Node]) -> None:
        """Saves a list of nodes to the database."""
        for i in range(len(nodes)):
            nodes[i]._id = self.save_node(nodes[i])._id

    def save_node_with_id(self, node: Node) -> Optional[Node]:
        """Saves a node to the database using the internal id."""
        results = self.execute_and_fetch(
            f"MATCH (node: {node._label})"
            f" WHERE id(node) = {node._id}"
            f" {node._get_cypher_set_properties('node')}"
            f" RETURN node;"
        )

        return self.get_variable_assume_one(results, "node")

    @abstractmethod
    def load_node(self, node: Node) -> Optional[Node]:
        """Loads a node from the database.
        If the node._id is not None, it fetches the node from the database with that
        internal id.
        If the node has unique fields it fetches the node from the database with
        those unique fields set.
        Otherwise it tries to find any node in the database that has all properties
        set to exactly the same values.
        If no node is found or no properties are set it raises a GQLAlchemyError.
        """
        pass

    def load_node_with_all_properties(self, node: Node) -> Optional[Node]:
        """Loads a node from the database with all equal property values."""
        results = self.execute_and_fetch(
            f"MATCH (node: {node._label}) WHERE {node._get_cypher_fields_and_block('node')} RETURN node;"
        )
        return self.get_variable_assume_one(results, "node")

    def load_node_with_id(self, node: Node) -> Optional[Node]:
        """Loads a node with the same internal database id."""
        results = self.execute_and_fetch(f"MATCH (node: {node._label}) WHERE id(node) = {node._id} RETURN node;")

        return self.get_variable_assume_one(results, "node")

    @abstractmethod
    def load_relationship(self, relationship: Relationship) -> Optional[Relationship]:
        """Returns a relationship loaded from the database.
        If the relationship._id is not None, it fetches the relationship from
        the database that has the same internal id.
        Otherwise it returns the relationship whose relationship._start_node_id
        and relationship._end_node_id and all relationship properties that
        are not None match the relationship in the database.
        If there is no relationship like that in database, or if there are
        multiple relationships like that in database, throws GQLAlchemyError.
        """
        pass

    def load_relationship_with_id(self, relationship: Relationship) -> Optional[Relationship]:
        """Loads a relationship from the database using the internal id."""
        results = self.execute_and_fetch(
            f"MATCH (start_node)-[relationship: {relationship._type}]->(end_node)"
            f" WHERE id(start_node) = {relationship._start_node_id}"
            f" AND id(end_node) = {relationship._end_node_id}"
            f" AND id(relationship) = {relationship._id}"
            f" RETURN relationship;"
        )
        return self.get_variable_assume_one(results, "relationship")

    def load_relationship_with_start_node_id_and_end_node_id(
        self, relationship: Relationship
    ) -> Optional[Relationship]:
        """Loads a relationship from the database using start node and end node id
        for which all properties of the relationship that are not None match.
        """
        and_block = relationship._get_cypher_fields_and_block("relationship")
        if and_block.strip():
            and_block = f" AND {and_block}"
        results = self.execute_and_fetch(
            f"MATCH (start_node)-[relationship:{relationship._type}]->(end_node)"
            f" WHERE id(start_node) = {relationship._start_node_id}"
            f" AND id(end_node) = {relationship._end_node_id}"
            f"{and_block} RETURN relationship;"
        )
        return self.get_variable_assume_one(results, "relationship")

    @abstractmethod
    def save_relationship(self, relationship: Relationship) -> Optional[Relationship]:
        """Saves a relationship to the database.
        If relationship._id is not None it finds the relationship in the database
        and updates it's properties with the values in `relationship`.
        If relationship._id is None, it creates a new relationship.
        If you want to set a relationship._id instead of creating a new
        relationship, use `load_relationship` first.
        """
        pass

    def save_relationships(self, relationships: List[Relationship]) -> None:
        """Saves a list of relationships to the database."""
        for i in range(len(relationships)):
            relationships[i]._id = self.save_relationship(relationships[i])._id

    def save_relationship_with_id(self, relationship: Relationship) -> Optional[Relationship]:
        """Saves a relationship to the database using the relationship._id."""
        results = self.execute_and_fetch(
            f"MATCH (start_node)-[relationship: {relationship._type}]->(end_node)"
            f" WHERE id(start_node) = {relationship._start_node_id}"
            f" AND id(end_node) = {relationship._end_node_id}"
            f" AND id(relationship) = {relationship._id}"
            f"{relationship._get_cypher_set_properties('relationship')} RETURN relationship;"
        )

        return self.get_variable_assume_one(results, "relationship")

    def create_relationship(self, relationship: Relationship) -> Optional[Relationship]:
        """Creates a new relationship in the database."""
        results = self.execute_and_fetch(
            "MATCH (start_node), (end_node)"
            f" WHERE id(start_node) = {relationship._start_node_id}"
            f" AND id(end_node) = {relationship._end_node_id}"
            f" CREATE (start_node)-[relationship:{relationship._type}]->(end_node)"
            f"{relationship._get_cypher_set_properties('relationship')} RETURN relationship"
        )

        return self.get_variable_assume_one(results, "relationship")



import os

MG_HOST = os.getenv("MG_HOST", "127.0.0.1")
MG_PORT = int(os.getenv("MG_PORT", "7687"))
MG_USERNAME = os.getenv("MG_USERNAME", "")
MG_PASSWORD = os.getenv("MG_PASSWORD", "")
MG_ENCRYPTED = os.getenv("MG_ENCRYPT", "false").lower() == "true"
MG_CLIENT_NAME = os.getenv("MG_CLIENT_NAME", "GQLAlchemy")
MG_LAZY = os.getenv("MG_LAZY", "false").lower() == "true"



# Copyright (c) 2016-2023 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Callable, List, Set, Dict, Tuple
from collections import defaultdict
from numbers import Number

try:
    import torch
except ModuleNotFoundError:
    torch = None

from gqlalchemy.exceptions import raise_if_not_imported
from gqlalchemy.transformations.constants import LABELS_CONCAT, DEFAULT_NODE_LABEL, DEFAULT_EDGE_TYPE
from gqlalchemy.memgraph_constants import (
    MG_HOST,
    MG_PORT,
    MG_USERNAME,
    MG_PASSWORD,
    MG_ENCRYPTED,
    MG_CLIENT_NAME,
    MG_LAZY,
)
from gqlalchemy.models import Node, Relationship
from gqlalchemy.utilities import to_cypher_properties
from gqlalchemy import Memgraph, Match


class Translator(ABC):
    # Lambda function to concat list of labels
    merge_labels: Callable[[Set[str]], str] = (
        lambda labels, default_node_label: LABELS_CONCAT.join([label for label in sorted(labels)])
        if len(labels)
        else default_node_label
    )

    @abstractmethod
    def __init__(
        self,
        host: str = MG_HOST,
        port: int = MG_PORT,
        username: str = MG_USERNAME,
        password: str = MG_PASSWORD,
        encrypted: bool = MG_ENCRYPTED,
        client_name: str = MG_CLIENT_NAME,
        lazy: bool = MG_LAZY,
    ) -> None:
        super().__init__()
        self.connection = Memgraph(host, port, username, password, encrypted, client_name, lazy)

    @abstractmethod
    def to_cypher_queries(graph):
        """Abstract method which doesn't know how to produce cypher queries for a specific graph type and thus needs to be overridden.
        Args:
            graph: Can be of any type supported by the derived Translator object.

        Raises:
            NotImplementedError: The method must be override by a specific translator.
        """
        raise NotImplementedError(
            "Subclasses must override this method to produce cypher queries for specific graph type."
        )

    @abstractmethod
    def get_instance():
        """Abstract method which doesn't know how to create the concrete instance so it needs to be overridden.

        Raises:
            NotImplementedError: The method must be override by a specific translator.
        """
        raise NotImplementedError(
            "Subclasses must override this method to correctly parse query results for specific graph type."
        )

    @classmethod
    def _is_most_inner_type_number(cls, obj):
        """Checks if the most inner object type is a type derived from Number. E.g For a List[List[int]] return True
        Args:
            obj: Object that will be checked.

        Returns:
            None if the object is empty(type cannot be inferred) or the type of the most inner object.
        """
        if isinstance(obj, list):
            return len(obj) and Translator._is_most_inner_type_number(obj[0])
        # Try to parser string into number
        return isinstance(obj, Number)

    @classmethod
    def validate_features(cls, features: List, expected_num: int):
        """Return true if features are okay to be set on all nodes/features.
        Args:
            features: To be set on all nodes. It can be anything that can be converted to torch tensor.
            expected_num: This can be number of nodes or number of edges depending on whether features will be set on nodes or edges.
        Returns:
            None if features cannot be set or tensor of same features.
        """
        raise_if_not_imported(dependency=torch, dependency_name="torch")

        if len(features) != expected_num:
            return None
        try:
            return torch.tensor(features, dtype=torch.float32)
        except ValueError:
            return None

    @classmethod
    def get_properties(cls, properties, entity_id):
        properties_ret = {}
        for property_key, property_values in properties.items():
            properties_ret[property_key] = property_values[entity_id]
        return properties_ret

    def _parse_memgraph(self):
        """Returns Memgraph's data parsed in a nice way to be used for creating DGL and PyG graphs."""
        mem_indexes: Dict[str, int] = defaultdict(int)  # track current counter for the node type
        node_features: Dict[str, Dict[str, List[int]]] = defaultdict(
            lambda: defaultdict(list)
        )  # node_label to prop_name to list of features for all nodes of that label
        edge_features: Dict[Tuple[str, str, str], Dict[str, List[int]]] = defaultdict(
            lambda: defaultdict(list)
        )  # (source_node_label, edge_type, dest_node_label) to prop_name to list of features for all edges of that type
        src_nodes: Dict[Tuple[str, str, str], List[int]] = defaultdict(
            list
        )  # key=(source_node_label, edge_type, destination_node_label), value=list of source nodes
        dest_nodes: Dict[Tuple[str, str, str], List[int]] = defaultdict(
            list
        )  # key=(source_node_label, edge_type, destination_node_label), value=list of dest nodes
        reindex: Dict[str, Dict[int, int]] = defaultdict(
            dict
        )  # Reindex structure for saving node_label to old_index to new_index

        rel_results = self.get_all_edges_from_db()

        for row in rel_results:
            row_values = row.values()
            for entity in row_values:
                entity_num_features = dict(
                    filter(lambda pair: Translator._is_most_inner_type_number(pair[1]), entity._properties.items())
                )
                if isinstance(entity, Node):
                    # Extract node label and all numeric features
                    node_label = Translator.merge_labels(entity._labels, DEFAULT_NODE_LABEL)
                    # Index node to a new index
                    if entity._id not in reindex[node_label]:
                        reindex[node_label][entity._id] = mem_indexes[node_label]
                        mem_indexes[node_label] += 1
                        # Copy numeric features of a node
                        for num_feature_key, num_feature_val in entity_num_features.items():
                            node_features[node_label][num_feature_key].append(num_feature_val)  # this should fail
                elif isinstance(entity, Relationship):
                    # Extract edge type and all numeric features
                    edge_type = entity._type if entity._type else DEFAULT_EDGE_TYPE
                    # Find descriptions of source and destination nodes. Cheap because we search only over row values, this is better than querying the database again.
                    source_node_index, dest_node_index = None, None
                    source_node_label, dest_node_label = None, None
                    for new_entity in row_values:
                        if new_entity._id == entity._start_node_id and isinstance(new_entity, Node):
                            source_node_label = Translator.merge_labels(new_entity._labels, DEFAULT_NODE_LABEL)
                            source_node_index = reindex[source_node_label][new_entity._id]
                        elif new_entity._id == entity._end_node_id and isinstance(new_entity, Node):
                            dest_node_label = Translator.merge_labels(new_entity._labels, DEFAULT_NODE_LABEL)
                            dest_node_index = reindex[dest_node_label][new_entity._id]
                    # Append to the graph data
                    edge_triplet = (source_node_label, edge_type, dest_node_label)
                    src_nodes[edge_triplet].append(source_node_index)
                    dest_nodes[edge_triplet].append(dest_node_index)
                    # Save edge features
                    for num_feature_key, num_feature_val in entity_num_features.items():
                        edge_features[edge_triplet][num_feature_key].append(num_feature_val)

        return src_nodes, dest_nodes, node_features, edge_features, mem_indexes

    def create_insert_query(
        self,
        source_node_label,
        source_node_properties,
        edge_type,
        edge_properties,
        dest_node_label,
        dest_node_properties,
    ):
        return (
            f"MERGE (n:{source_node_label} {to_cypher_properties(source_node_properties)}) "
            f"MERGE (m:{dest_node_label} {to_cypher_properties(dest_node_properties)}) "
            f"MERGE (n)-[r:{edge_type} {to_cypher_properties(edge_properties)}]->(m)"
        )

    def get_all_edges_from_db(self):
        """Returns all edges from the database.
        Returns:
            Query results when finding all edges.
        """
        return (
            Match(connection=self.connection).node(variable="n").to(variable="r").node(variable="m").return_().execute()
        )

    def get_all_isolated_nodes_from_db(self):
        """Returns all isolated nodes from the database.
        Returns:
            Query results for finding all isolated nodes.
        """
        return (
            Match(connection=self.connection)
            .node(variable="n")
            .add_custom_cypher("WHERE degree(n) = 0")
            .return_()
            .execute()
        )



# Copyright (c) 2016-2023 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterator, List, Dict, Any, Union
import logging
import multiprocessing as mp

import networkx as nx
import mgclient


from gqlalchemy import Memgraph
from gqlalchemy.transformations.constants import LABEL, EDGE_TYPE, DEFAULT_EDGE_TYPE, DEFAULT_NODE_LABEL
from gqlalchemy.transformations.translators.translator import Translator
from gqlalchemy.models import Node, Relationship, MemgraphIndex
from gqlalchemy.utilities import NetworkXCypherConfig, to_cypher_labels, to_cypher_properties, to_cypher_value
from gqlalchemy.memgraph_constants import (
    MG_HOST,
    MG_PORT,
    MG_USERNAME,
    MG_PASSWORD,
    MG_ENCRYPTED,
    MG_CLIENT_NAME,
    MG_LAZY,
)


class NetworkXGraphConstants:
    LABELS = "labels"
    TYPE = "type"
    ID = "id"


class NetworkXCypherBuilder:
    def __init__(self, config: NetworkXCypherConfig):
        if config is None:
            raise NoNetworkXConfigException("No NetworkX configuration provided!")

        self._config = config

    def yield_queries(self, graph: nx.Graph) -> Iterator[str]:
        """Generates Cypher queries for creating a graph."""
        if graph is None:
            return []

        if self._config.create_index:
            yield from self._nx_nodes_to_cypher_with_index(graph)
        else:
            yield from self._nx_nodes_to_cypher(graph)
        yield from self._nx_edges_to_cypher(graph)

    def yield_query_groups(self, graph: nx.Graph) -> List[Iterator[str]]:
        """Generates Cypher queries for creating a graph by query groups."""

        query_groups = []

        if self._config.create_index:
            query_groups.append(self._nx_nodes_to_cypher_with_index(graph))
        else:
            query_groups.append(self._nx_nodes_to_cypher(graph))

        query_groups.append(self._nx_edges_to_cypher(graph))

        return query_groups

    def _nx_nodes_to_cypher(self, graph: nx.Graph) -> Iterator[str]:
        """Generates Cypher queries for creating nodes."""
        for nx_id, data in graph.nodes(data=True):
            yield self._create_node(nx_id, data)

    def _nx_nodes_to_cypher_with_index(self, graph: nx.Graph) -> Iterator[str]:
        """Generates Cypher queries for creating nodes and indexes."""
        labels = set()
        for nx_id, data in graph.nodes(data=True):
            node_labels = data.get(NetworkXGraphConstants.LABELS, None)
            if isinstance(node_labels, (list, set)):
                labels |= set(node_labels)
            else:
                labels.add(node_labels)
            yield self._create_node(nx_id, data)
        labels.discard(None)
        for label in labels:
            yield self._create_index(label)

    def _nx_edges_to_cypher(self, graph: nx.Graph) -> Iterator[str]:
        """Generates Cypher queries for creating edges."""
        for n1, n2, data in graph.edges(data=True):
            from_label = graph.nodes[n1].get(NetworkXGraphConstants.LABELS, "")
            to_label = graph.nodes[n2].get(NetworkXGraphConstants.LABELS, "")

            n1_id = graph.nodes[n1].get(NetworkXGraphConstants.ID, n1)
            n2_id = graph.nodes[n2].get(NetworkXGraphConstants.ID, n2)
            yield self._create_edge(n1_id, n2_id, from_label, to_label, data)

    def _create_node(self, nx_id: int, properties: Dict[str, Any]) -> str:
        """Returns Cypher query for node creation."""
        if "id" not in properties:
            properties["id"] = nx_id
        labels_str = to_cypher_labels(properties.get(NetworkXGraphConstants.LABELS, ""))
        properties_without_labels = {k: v for k, v in properties.items() if k != NetworkXGraphConstants.LABELS}
        properties_str = to_cypher_properties(properties_without_labels, self._config)

        return f"CREATE ({labels_str} {properties_str});"

    def _create_edge(
        self,
        from_id: Union[int, str],
        to_id: Union[int, str],
        from_label: Union[str, List[str]],
        to_label: Union[str, List[str]],
        properties: Dict[str, Any],
    ) -> str:
        """Returns Cypher query for edge creation."""
        edge_type = to_cypher_labels(properties.get(NetworkXGraphConstants.TYPE, "TO"))
        properties.pop(NetworkXGraphConstants.TYPE, None)
        properties_str = to_cypher_properties(properties, self._config)
        from_label_str = to_cypher_labels(from_label)
        to_label_str = to_cypher_labels(to_label)
        from_id_str = to_cypher_value(from_id)
        to_id_str = to_cypher_value(to_id)

        return f"MATCH (n{from_label_str} {{id: {from_id_str}}}), (m{to_label_str} {{id: {to_id_str}}}) CREATE (n)-[{edge_type} {properties_str}]->(m);"

    def _create_index(self, label: str, property: str = None):
        """Returns Cypher query for index creation."""
        index = MemgraphIndex(label, property)
        return f"CREATE INDEX ON {index.to_cypher()}(id);"


class NoNetworkXConfigException(Exception):
    pass


class NxTranslator(Translator):
    """Uses original ids from Memgraph. Labels are encoded as properties. Since NetworkX allows
    that nodes have properties of different dimensionality, this modules makes use of it and stores properties
    as dictionary entries. All properties are saved to NetworkX data structure.
    """

    def __init__(
        self,
        host: str = MG_HOST,
        port: int = MG_PORT,
        username: str = MG_USERNAME,
        password: str = MG_PASSWORD,
        encrypted: bool = MG_ENCRYPTED,
        client_name: str = MG_CLIENT_NAME,
        lazy: bool = MG_LAZY,
    ) -> None:
        self.__all__ = ("nx_to_cypher", "nx_graph_to_memgraph_parallel")
        super().__init__(host, port, username, password, encrypted, client_name, lazy)

    def to_cypher_queries(self, graph: nx.Graph, config: NetworkXCypherConfig = None) -> Iterator[str]:
        """Generates a Cypher query for creating a graph."""

        if config is None:
            config = NetworkXCypherConfig()

        builder = NetworkXCypherBuilder(config=config)

        yield from builder.yield_queries(graph)

    def nx_graph_to_memgraph_parallel(
        self,
        graph: nx.Graph,
        config: NetworkXCypherConfig = None,
    ) -> None:
        """Generates Cypher queries and inserts data into Memgraph in parallel."""
        if config is None:
            config = NetworkXCypherConfig()

        builder = NetworkXCypherBuilder(config=config)
        query_groups = builder.yield_query_groups(graph)

        if not config.create_index:
            self._check_for_index_hint(
                self.host,
                self.port,
                self.username,
                self.password,
                self.encrypted,
            )

        for query_group in query_groups:
            self._start_parallel_execution(
                query_group, self.host, self.port, self.username, self.password, self.encrypted
            )

    def _start_parallel_execution(self, queries_gen: Iterator[str]) -> None:
        num_of_processes = mp.cpu_count() // 2
        queries = list(queries_gen)
        chunk_size = len(queries) // num_of_processes
        processes = []

        for i in range(num_of_processes):
            process_queries = queries[i * chunk_size : chunk_size * (i + 1)]
            processes.append(
                mp.Process(
                    target=self._insert_queries,
                    args=(
                        process_queries,
                        self.host,
                        self.port,
                        self.username,
                        self.password,
                        self.encrypted,
                    ),
                )
            )
        for p in processes:
            p.start()
        for p in processes:
            p.join()

    def _insert_queries(
        self, queries: List[str], host: str, port: int, username: str, password: str, encrypted: bool
    ) -> None:
        """Used by multiprocess insertion of nx into memgraph, works on a chunk of queries."""
        memgraph = Memgraph(host, port, username, password, encrypted)
        while len(queries) > 0:
            try:
                query = queries.pop()
                memgraph.execute(query)
            except mgclient.DatabaseError as e:
                queries.append(query)
                logging.getLogger(__file__).warning(f"Ignoring database error: {e}")
                continue

    def _check_for_index_hint(
        self,
        host: str = "127.0.0.1",
        port: int = 7687,
        username: str = "",
        password: str = "",
        encrypted: bool = False,
    ):
        """Check if the there are indexes, if not show warnings."""
        memgraph = Memgraph(host, port, username, password, encrypted)
        indexes = memgraph.get_indexes()
        if len(indexes) == 0:
            logging.getLogger(__file__).warning(
                "Be careful you do not have any indexes set up, the queries will take longer than expected!"
            )

    def get_instance(self):
        """Creates NetworkX instance of the graph from the data residing inside Memgraph. Since NetworkX doesn't support labels in a way Memgraph does, labels
        are encoded as a node and edge properties.
        """
        # Get all nodes and edges from the database (connected nodes)
        rel_results = self.get_all_edges_from_db()
        # Data structures
        graph_data = []  # List[Tuple[source_node_id, dest_node_id]]
        node_info = dict()  # Dict[id, Dict[prop: value]]
        edge_info = dict()  # Dict[(source_node_id, dest_node_id), Dict[prop: value]]

        # Parse each row from query results
        for row in rel_results:
            row_values = row.values()
            for entity in row_values:
                entity_properties = entity._properties
                if isinstance(entity, Node) and entity._id not in node_info:
                    entity_properties[LABEL] = Translator.merge_labels(entity._labels, DEFAULT_NODE_LABEL)
                    node_info[entity._id] = entity_properties
                elif isinstance(entity, Relationship):
                    entity_properties[EDGE_TYPE] = entity._type if entity._type else DEFAULT_EDGE_TYPE
                    edge_info[(entity._start_node_id, entity._end_node_id)] = entity_properties
                    graph_data.append((entity._start_node_id, entity._end_node_id))

        # Create Nx graph
        graph = nx.DiGraph(graph_data)
        nx.set_node_attributes(graph, node_info)
        nx.set_edge_attributes(graph, edge_info)
        # Process all isolated nodes
        isolated_nodes_results = self.get_all_isolated_nodes_from_db()
        for isolated_node in isolated_nodes_results:
            isolated_node = isolated_node["n"]  #
            entity_properties = isolated_node._properties
            entity_properties[LABEL] = Translator.merge_labels(isolated_node._labels, DEFAULT_NODE_LABEL)
            graph.add_node(isolated_node._id, **entity_properties)  # unpack dictionary

        return graph



LABELS_CONCAT = ":"
LABEL = "label"
EDGE_TYPE = "type"
DGL_ID = "dgl_id"
DGL_DEFAULT_NODE_LABEL = "_N"
DGL_DEFAULT_EDGE_TYPE = "_E"
PYG_ID = "pyg_id"
EDGE_INDEX = "edge_index"
NUM_NODES = "num_nodes"
DEFAULT_NODE_LABEL = "NODE"
DEFAULT_EDGE_TYPE = "RELATIONSHIP"



# Copyright (c) 2016-2022 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod

BFS_EXPANSION = " *BFS"
DFS_EXPANSION = " *"
WSHORTEST_EXPANSION = " *WSHORTEST"
ALLSHORTEST_EXPANSION = " *ALLSHORTEST"

DEFAULT_TOTAL_WEIGHT = "total_weight"
DEFAULT_WEIGHT_PROPERTY = "r.weight"


class IntegratedAlgorithm(ABC):
    """Abstract class modeling Memgraph's built-in graph algorithms.

    These algorithms are integrated into Memgraph's codebase and are called
    within a relationship part of a query. For instance:
    MATCH p = (:City {name: "Paris"})
          -[:Road * bfs (r, n | r.length <= 200 AND n.name != "Metz")]->
          (:City {name: "Berlin"})
    """

    @abstractmethod
    def __str__(self) -> str:
        """Instance of IntegratedAlgorithm subclass is used as a string"""
        pass

    @staticmethod
    def to_cypher_lambda(expression: str) -> str:
        """Method for creating a general lambda expression.

        Variables `r` and `n` stand for relationship and node. The expression is
        used e.g. for a filter lambda, to use only relationships of length less
        than 200:
            expression="r.length < 200"
        with the filter lambda being:
            (r, n | r.length < 200)

        Args:
            expression: Lambda conditions or statements.
        """
        return "" if expression is None else f"(r, n | {expression})"


class BreadthFirstSearch(IntegratedAlgorithm):
    """Build a BFS call for a Cypher query.

    The Breadth-first search can be called in Memgraph with Cypher queries such
    as: `MATCH (a {id: 723})-[*BFS ..10 (r, n | r.x > 12 AND n.y < 3)]-() RETURN *;`
    It is called inside the relationship clause, `*BFS` naming the algorithm,
    `..10` specifying depth bounds, and `(r, n | <expression>)` is a filter
    lambda.
    """

    def __init__(
        self,
        lower_bound: int = None,
        upper_bound: int = None,
        condition: str = None,
    ) -> None:
        """
        Args:
            lower_bound: Lower bound for path depth.
            upper_bound: Upper bound for path depth.
            condition: Filter through nodes and relationships that pass this
            condition.
        """
        super().__init__()
        self.lower_bound = str(lower_bound) if lower_bound is not None else ""
        self.upper_bound = str(upper_bound) if upper_bound is not None else ""
        self.condition = condition

    def __str__(self) -> str:
        """Get a Cypher query string for this algorithm."""
        algo_str = BFS_EXPANSION

        bounds = self.to_cypher_bounds()
        if bounds != "":
            algo_str = f"{algo_str} {bounds}"

        filter_lambda = super().to_cypher_lambda(self.condition)
        if filter_lambda != "":
            algo_str = f"{algo_str} {filter_lambda}"

        return algo_str

    def to_cypher_bounds(self) -> str:
        """If bounds are specified, returns them in grammar-defined form."""
        if self.lower_bound == "" and self.upper_bound == "":
            return ""

        return f"{self.lower_bound}..{self.upper_bound}"


class DepthFirstSearch(IntegratedAlgorithm):
    """Build a DFS call for a Cypher query.
    The Depth-First Search can be called in Memgraph with Cypher queries
    such as:
    MATCH (a {id: 723})-[* ..10 (r, n | r.x > 12 AND n.y < 3)]-() RETURN *;
    It is called inside the relationship clause, "*" naming the algorithm
    ("*" without "DFS" because it is defined like such in openCypher),
    "..10" specifying depth bounds, and "(r, n | <expression>)" is a filter
    lambda.
    """

    def __init__(
        self,
        lower_bound: int = None,
        upper_bound: int = None,
        condition: str = None,
    ) -> None:
        """
        Args:
            lower_bound: Lower bound for path depth.
            upper_bound: Upper bound for path depth.
            condition: Filter through nodes and relationships that pass this
            condition.
        """
        super().__init__()
        self.lower_bound = str(lower_bound) if lower_bound is not None else ""
        self.upper_bound = str(upper_bound) if upper_bound is not None else ""
        self.condition = condition

    def __str__(self) -> str:
        """get Cypher query string for this algorithm."""
        algo_str = DFS_EXPANSION

        bounds = self.to_cypher_bounds()
        if bounds != "":
            algo_str = f"{algo_str} {bounds}"

        filter_lambda = super().to_cypher_lambda(self.condition)
        if filter_lambda != "":
            algo_str = f"{algo_str} {filter_lambda}"

        return algo_str

    def to_cypher_bounds(self) -> str:
        """If bounds are specified, returns them in grammar-defined form."""
        if self.lower_bound == "" and self.upper_bound == "":
            return ""

        return f"{self.lower_bound}..{self.upper_bound}"


class WeightedShortestPath(IntegratedAlgorithm):
    """Build a Dijkstra shortest path call for a Cypher query.

    The weighted shortest path algorithm can be called in Memgraph with Cypher
    queries such as:
    " MATCH (a {id: 723})-[r *WSHORTEST 10 (r, n | r.weight) weight_sum
            (r, n | r.x > 12 AND r.y < 3)]-(b {id: 882}) RETURN * "
    It is called inside the relationship clause, "*WSHORTEST" naming the
    algorithm, "10" specifying search depth bounds, and "(r, n | <expression>)"
    is a filter lambda, used to filter which relationships and nodes to use.
    """

    def __init__(
        self,
        upper_bound: int = None,
        condition: str = None,
        total_weight_var: str = DEFAULT_TOTAL_WEIGHT,
        weight_property: str = DEFAULT_WEIGHT_PROPERTY,
    ) -> None:
        """
        Args:
            upper_bound: Upper bound for path depth.
            condition: Filter through nodes and relationships that pass this
            condition.
            total_weight_var: Variable defined as the sum of all weights on
            path being returned.
            weight_property: property being used as weight.
        """
        super().__init__()
        self.weight_property = weight_property if "." in weight_property else f"r.{weight_property}"
        self.total_weight_var = total_weight_var
        self.condition = condition
        self.upper_bound = "" if upper_bound is None else str(upper_bound)

    def __str__(self) -> str:
        algo_str = WSHORTEST_EXPANSION
        if self.upper_bound != "":
            algo_str = f"{algo_str} {self.upper_bound}"

        algo_str = f"{algo_str} {super().to_cypher_lambda(self.weight_property)} {self.total_weight_var}"

        filter_lambda = super().to_cypher_lambda(self.condition)
        if filter_lambda != "":
            algo_str = f"{algo_str} {filter_lambda}"

        return algo_str


class AllShortestPath(IntegratedAlgorithm):
    """Build a Dijkstra shortest path call for a Cypher query.

    The weighted shortest path algorithm can be called in Memgraph with Cypher
    queries such as:
    " MATCH (a {id: 723})-[r *ALLSHORTEST 10 (r, n | r.weight) total_weight
            (r, n | r.x > 12 AND r.y < 3)]-(b {id: 882}) RETURN * "
    It is called inside the relationship clause, "*ALLSHORTEST" naming the
    algorithm, "10" specifying search depth bounds, and "(r, n | <expression>)"
    is a filter lambda, used to filter which relationships and nodes to use.
    """

    def __init__(
        self,
        upper_bound: int = None,
        condition: str = None,
        total_weight_var: str = DEFAULT_TOTAL_WEIGHT,
        weight_property: str = DEFAULT_WEIGHT_PROPERTY,
    ) -> None:
        """
        Args:
            upper_bound: Upper bound for path depth.
            condition: Filter through nodes and relationships that pass this
            condition.
            total_weight_var: Variable defined as the sum of all weights on
            path being returned.
            weight_property: Property being used as weight.
        """
        super().__init__()
        self.weight_property = weight_property if "." in weight_property else f"r.{weight_property}"
        self.total_weight_var = total_weight_var
        self.condition = condition
        self.upper_bound = "" if upper_bound is None else str(upper_bound)

    def __str__(self) -> str:
        algo_str = ALLSHORTEST_EXPANSION
        if self.upper_bound != "":
            algo_str = f"{algo_str} {self.upper_bound}"

        algo_str = f"{algo_str} {super().to_cypher_lambda(self.weight_property)} {self.total_weight_var}"

        filter_lambda = super().to_cypher_lambda(self.condition)
        if filter_lambda != "":
            algo_str = f"{algo_str} {filter_lambda}"

        return algo_str



# Copyright (c) 2016-2022 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Dict

import mgp
from kafka.producer import KafkaProducer


producers_by_name: Dict[str, KafkaProducer] = {}
topics_by_name: Dict[str, str] = {}


@mgp.read_proc
def create_push_stream(
    context: mgp.ProcCtx,
    stream_name: str,
    topic: str,
    config: mgp.Map,
) -> mgp.Record(created=bool):

    if not isinstance(stream_name, str):
        raise TypeError("Invalid type on first argument!")

    if not isinstance(topic, str):
        raise TypeError("Invalid type on second argument!")

    producer = KafkaProducer(value_serializer=lambda m: json.dumps(m).encode("utf-8"), **config)

    producers_by_name[stream_name] = producer
    topics_by_name[stream_name] = topic

    return mgp.Record(created=True)


@mgp.read_proc
def show_streams(
    context: mgp.ProcCtx,
) -> mgp.Record(name=str, topic=str):
    records = []

    for k, v in topics_by_name.items():
        records.append(mgp.Record(name=k, topic=v))

    return records


@mgp.read_proc
def push(
    context: mgp.ProcCtx,
    stream_name: str,
    payload: mgp.Map,
) -> mgp.Record(message=mgp.Map):

    if not isinstance(stream_name, str):
        raise TypeError("Invalid type on first argument!")

    if not isinstance(payload, mgp.Map):
        raise TypeError("Invalid type on second argument!")

    if stream_name not in producers_by_name or stream_name not in topics_by_name:
        raise Exception(f"Stream {stream_name} is not present!")

    message = ""
    if isinstance(payload, dict):
        message = payload
    elif isinstance(payload, mgp.Vertex) or isinstance(payload, mgp.Edge):
        message = {x.name: x.value for x in payload.properties.items()}
    else:
        raise Exception("Can't have message type other than Map / Vertex / Edge")

    producer, topic = producers_by_name[stream_name], topics_by_name[stream_name]

    try:
        producer.send(topic, message)
    except Exception as e:
        raise Exception(f"Exception when sending message: {e}")

    return mgp.Record(message=message)



# Copyright (c) 2016-2022 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
import os
import sqlite3
from typing import List, Optional, Union

from gqlalchemy.connection import Connection, MemgraphConnection
from gqlalchemy.disk_storage import OnDiskPropertyDatabase
from gqlalchemy.exceptions import (
    GQLAlchemyError,
    GQLAlchemyFileNotFoundError,
    GQLAlchemyOnDiskPropertyDatabaseNotDefinedError,
    GQLAlchemyUniquenessConstraintError,
)
from gqlalchemy.models import (
    MemgraphConstraintExists,
    MemgraphConstraintUnique,
    MemgraphIndex,
    MemgraphStream,
    MemgraphTrigger,
    Node,
    Relationship,
)
from gqlalchemy.vendors.database_client import DatabaseClient
from gqlalchemy.graph_algorithms.query_modules import QueryModule
import gqlalchemy.memgraph_constants as mg_consts

__all__ = ("Memgraph",)


class MemgraphTransaction:
    def __init__(self, username: str, transaction_id: str, query: list, metadata: dict):
        self.username = (username,)
        self.transaction_id = transaction_id
        self.query = query
        self.metadata = metadata

    def __repr__(self):
        return f"MemgraphTransaction(username={self.username}, transaction_id={self.transaction_id}, query={self.query}, metadata={self.metadata})"


class MemgraphTerminatedTransaction:
    def __init__(self, transaction_id: str, killed: bool):
        self.transaction_id = transaction_id
        self.killed = killed

    def __repr__(self):
        return f"MemgraphTerminatedTransaction(transaction_id={self.transaction_id}, killed={self.killed})"


class MemgraphConstants:
    CONSTRAINT_TYPE = "constraint type"
    EXISTS = "exists"
    LABEL = "label"
    PROPERTY = "property"
    PROPERTIES = "properties"
    UNIQUE = "unique"


class MemgraphStorageMode(Enum):
    IN_MEMORY_TRANSACTIONAL = "IN_MEMORY_TRANSACTIONAL"
    IN_MEMORY_ANALYTICAL = "IN_MEMORY_ANALYTICAL"
    ON_DISK_TRANSACTIONAL = "ON_DISK_TRANSACTIONAL"

    def __str__(self):
        return self.value


def create_transaction(transaction_data) -> MemgraphTransaction:
    """Create a MemgraphTransaction object from transaction data.
    Args:
        transaction_data (dict): A dictionary containing transaction data.
    Returns:
        MemgraphTransaction: A MemgraphTransaction object.
    """
    return MemgraphTransaction(
        username=transaction_data["username"],
        transaction_id=transaction_data["transaction_id"],
        query=transaction_data["query"],
        metadata=transaction_data["metadata"],
    )


def create_terminated_transaction(transaction_data) -> MemgraphTerminatedTransaction:
    """Create a MemgraphTerminatedTransaction object from transaction data.
    Args:
        transaction_data (dict): A dictionary containing transaction data.
    Returns:
        MemgraphTerminatedTransaction: A MemgraphTerminatedTransaction object.
    """
    return MemgraphTerminatedTransaction(
        transaction_id=transaction_data["transaction_id"],
        killed=transaction_data["killed"],
    )


class Memgraph(DatabaseClient):
    def __init__(
        self,
        host: str = mg_consts.MG_HOST,
        port: int = mg_consts.MG_PORT,
        username: str = mg_consts.MG_USERNAME,
        password: str = mg_consts.MG_PASSWORD,
        encrypted: bool = mg_consts.MG_ENCRYPTED,
        client_name: str = mg_consts.MG_CLIENT_NAME,
        lazy: bool = mg_consts.MG_LAZY,
    ):
        super().__init__(
            host=host, port=port, username=username, password=password, encrypted=encrypted, client_name=client_name
        )
        self._lazy = lazy
        self._on_disk_db = None

    def get_indexes(self) -> List[MemgraphIndex]:
        """Returns a list of all database indexes (label and label-property types)."""
        indexes = []
        for result in self.execute_and_fetch("SHOW INDEX INFO;"):
            indexes.append(
                MemgraphIndex(
                    result[MemgraphConstants.LABEL],
                    result[MemgraphConstants.PROPERTY],
                )
            )
        return indexes

    def ensure_indexes(self, indexes: List[MemgraphIndex]) -> None:
        """Ensures that database indexes match input indexes."""
        old_indexes = set(self.get_indexes())
        new_indexes = set(indexes)
        for obsolete_index in old_indexes.difference(new_indexes):
            self.drop_index(obsolete_index)
        for missing_index in new_indexes.difference(old_indexes):
            self.create_index(missing_index)

    def get_constraints(
        self,
    ) -> List[Union[MemgraphConstraintExists, MemgraphConstraintUnique]]:
        """Returns a list of all database constraints (label and label-property types)."""
        constraints: List[Union[MemgraphConstraintExists, MemgraphConstraintUnique]] = []
        for result in self.execute_and_fetch("SHOW CONSTRAINT INFO;"):
            if result[MemgraphConstants.CONSTRAINT_TYPE] == MemgraphConstants.UNIQUE:
                constraints.append(
                    MemgraphConstraintUnique(
                        result[MemgraphConstants.LABEL],
                        tuple(result[MemgraphConstants.PROPERTIES]),
                    )
                )
            elif result[MemgraphConstants.CONSTRAINT_TYPE] == MemgraphConstants.EXISTS:
                constraints.append(
                    MemgraphConstraintExists(
                        result[MemgraphConstants.LABEL],
                        result[MemgraphConstants.PROPERTIES],
                    )
                )
        return constraints

    def get_exists_constraints(
        self,
    ) -> List[MemgraphConstraintExists]:
        return [x for x in self.get_constraints() if isinstance(x, MemgraphConstraintExists)]

    def get_unique_constraints(
        self,
    ) -> List[MemgraphConstraintUnique]:
        return [x for x in self.get_constraints() if isinstance(x, MemgraphConstraintUnique)]

    def new_connection(self) -> Connection:
        """Creates new Memgraph connection."""
        args = dict(
            host=self._host,
            port=self._port,
            username=self._username,
            password=self._password,
            encrypted=self._encrypted,
            client_name=self._client_name,
        )
        return MemgraphConnection(**args)

    def create_stream(self, stream: MemgraphStream) -> None:
        """Create a stream."""
        query = stream.to_cypher()
        self.execute(query)

    def start_stream(self, stream: MemgraphStream) -> None:
        """Start a stream."""
        query = f"START STREAM {stream.name};"
        self.execute(query)

    def get_streams(self) -> List[str]:
        """Returns a list of all streams."""
        streams = []
        for result in self.execute_and_fetch("SHOW STREAMS;"):
            streams.append(result)
        return streams

    def drop_stream(self, stream: MemgraphStream) -> None:
        """Drop a stream."""
        query = f"DROP STREAM {stream.name};"
        self.execute(query)

    def create_trigger(self, trigger: MemgraphTrigger) -> None:
        """Creates a trigger."""
        query = trigger.to_cypher()
        self.execute(query)

    def get_triggers(self) -> List[MemgraphTrigger]:
        """Returns a list of all database triggers."""
        triggers_list = list(self.execute_and_fetch("SHOW TRIGGERS;"))
        memgraph_triggers_list = []
        for trigger in triggers_list:
            event_type = trigger["event type"]
            event_object = None

            if event_type == "ANY":
                event_type = None
            elif len(event_type.split()) > 1:
                [event_object, event_type] = [part for part in event_type.split()]

            memgraph_triggers_list.append(
                MemgraphTrigger(
                    name=trigger["trigger name"],
                    event_type=event_type,
                    event_object=event_object,
                    execution_phase=trigger["phase"].split()[0],
                    statement=trigger["statement"],
                )
            )
        return memgraph_triggers_list

    def drop_trigger(self, trigger: MemgraphTrigger) -> None:
        """Drop a trigger."""
        query = f"DROP TRIGGER {trigger.name};"
        self.execute(query)

    def drop_triggers(self) -> None:
        """Drops all triggers in the database."""
        for trigger in self.get_triggers():
            self.drop_trigger(trigger)

    def _new_connection(self) -> Connection:
        """Creates new Memgraph connection."""
        args = dict(
            host=self._host,
            port=self._port,
            username=self._username,
            password=self._password,
            encrypted=self._encrypted,
            client_name=self._client_name,
        )
        return MemgraphConnection(**args)

    def init_disk_storage(self, on_disk_db: OnDiskPropertyDatabase) -> None:
        """Adds and OnDiskPropertyDatabase to the database so that any property
        that has a Field(on_disk=True) can be stored to and loaded from
        an OnDiskPropertyDatabase.
        """
        self.on_disk_db = on_disk_db

    def remove_on_disk_storage(self) -> None:
        """Removes the OnDiskPropertyDatabase from the database."""
        self.on_disk_db = None

    def save_node(self, node: Node) -> Node:
        """Saves node to the database.
        If the node._id is not None it fetches the node with the same id from
        the database and updates it's fields.
        If the node has unique fields it fetches the nodes with the same unique
        fields from the database and updates it's fields.
        Otherwise it creates a new node with the same properties.
        Null properties are ignored.
        """
        result = None
        if node._id is not None:
            result = self.save_node_with_id(node)
        elif node.has_unique_fields():
            matching_nodes = list(self._get_nodes_with_unique_fields(node))
            if len(matching_nodes) > 1:
                raise GQLAlchemyUniquenessConstraintError(
                    f"Uniqueness constraints match multiple nodes: {matching_nodes}"
                )
            elif len(matching_nodes) == 1:
                node._id = matching_nodes[0]["node"]._id
                result = self.save_node_with_id(node)
            else:
                result = self.create_node(node)
        else:
            result = self.create_node(node)

        result = self._save_node_properties_on_disk(node, result)
        return result

    def _save_node_properties_on_disk(self, node: Node, result: Node) -> Node:
        """Saves all on_disk properties to the on disk database attached to
        the database.
        """
        for field in node.__fields__:
            value = getattr(node, field, None)
            if value is not None and "on_disk" in node.__fields__[field].field_info.extra:
                if self.on_disk_db is None:
                    raise GQLAlchemyOnDiskPropertyDatabaseNotDefinedError()
                self.on_disk_db.save_node_property(result._id, field, value)
                setattr(result, field, value)

        return result

    def load_node(self, node: Node) -> Optional[Node]:
        """Loads a node from the database.
        If the node._id is not None it fetches the node from the database with that
        internal id.
        If the node has unique fields it fetches the node from the database with
        those unique fields set.
        Otherwise it tries to find any node in the database that has all properties
        set to exactly the same values.
        If no node is found or no properties are set it raises a GQLAlchemyError.
        """
        if node._id is not None:
            result = self.load_node_with_id(node)
        elif node.has_unique_fields():
            matching_node = self.get_variable_assume_one(
                query_result=self._get_nodes_with_unique_fields(node), variable_name="node"
            )
            result = matching_node
        else:
            result = self.load_node_with_all_properties(node)

        result = self._load_node_properties_on_disk(result)
        return result

    def _load_node_properties_on_disk(self, result: Node) -> Node:
        """Loads all on_disk properties from the on disk database."""
        for field in result.__fields__:
            value = getattr(result, field, None)
            if "on_disk" in result.__fields__[field].field_info.extra:
                if self.on_disk_db is None:
                    raise GQLAlchemyOnDiskPropertyDatabaseNotDefinedError()
                try:
                    new_value = self.on_disk_db.load_node_property(result._id, field)
                except sqlite3.OperationalError:
                    new_value = value
                setattr(result, field, new_value)

        return result

    def load_relationship(self, relationship: Relationship) -> Optional[Relationship]:
        """Returns a relationship loaded from the database.
        If the relationship._id is not None it fetches the relationship from
        the database that has the same internal id.
        Otherwise it returns the relationship whose relationship._start_node_id
        and relationship._end_node_id and all relationship properties that
        are not None match the relationship in the database.
        If there is no relationship like that in the database, or if there are
        multiple relationships like that in the database, throws GQLAlchemyError.
        """
        if relationship._id is not None:
            result = self.load_relationship_with_id(relationship)
        elif relationship._start_node_id is not None and relationship._end_node_id is not None:
            result = self.load_relationship_with_start_node_id_and_end_node_id(relationship)
        else:
            raise GQLAlchemyError("Can't load a relationship without a start_node_id and end_node_id.")
        result = self._load_relationship_properties_on_disk(result)
        return result

    def _load_relationship_properties_on_disk(self, result: Relationship) -> Relationship:
        """Returns the relationship with all on_disk properties loaded from
        the OnDiskPropertyDatabase.
        If there is no OnDiskPropertyDatabase set with
        Memgraph().init_disk_storage() throws a
        GQLAlchemyOnDiskPropertyDatabaseNotDefinedError.
        """
        for field in result.__fields__:
            value = getattr(result, field, None)
            if "on_disk" in result.__fields__[field].field_info.extra:
                if self.on_disk_db is None:
                    raise GQLAlchemyOnDiskPropertyDatabaseNotDefinedError()
                try:
                    new_value = self.on_disk_db.load_relationship_property(result._id, field)
                except sqlite3.OperationalError:
                    new_value = value
                setattr(result, field, new_value)

        return result

    def save_relationship(self, relationship: Relationship) -> Optional[Relationship]:
        """Saves a relationship to the database.
        If relationship._id is not None it finds the relationship in the database
        and updates it's properties with the values in `relationship`.
        If relationship._id is None, it creates a new relationship.
        If you want to set a relationship._id instead of creating a new
        relationship, use `load_relationship` first.
        """
        if relationship._id is not None:
            result = self.save_relationship_with_id(relationship)
        elif relationship._start_node_id is not None and relationship._end_node_id is not None:
            result = self.create_relationship(relationship)
        else:
            raise GQLAlchemyError("Can't create a relationship without start_node_id and end_node_id.")

        result = self._save_relationship_properties_on_disk(relationship, result)
        return result

    def _save_relationship_properties_on_disk(self, relationship: Relationship, result: Relationship) -> Relationship:
        """Saves on_disk relationship properties on the OnDiskPropertyDatabase
        added with Memgraph().init_disk_storage(db). If OnDiskPropertyDatabase
        is not defined raises GQLAlchemyOnDiskPropertyDatabaseNotDefinedError.
        """
        for field in relationship.__fields__:
            value = getattr(relationship, field, None)
            if value is not None and "on_disk" in relationship.__fields__[field].field_info.extra:
                if self.on_disk_db is None:
                    raise GQLAlchemyOnDiskPropertyDatabaseNotDefinedError()
                self.on_disk_db.save_relationship_property(result._id, field, value)
                setattr(result, field, value)

        return result

    def get_procedures(self, starts_with: Optional[str] = None, update: bool = False) -> List["QueryModule"]:
        """Return query procedures.

        Maintains a list of query modules in the Memgraph object. If starts_with
        is defined then return those modules that start with starts_with string.

        Args:
            starts_with: Return those modules that start with this string.
            (Optional)
            update: Whether to update the list of modules in
            self.query_modules. (Optional)
        """
        if not hasattr(self, "query_modules") or update:
            results = self.execute_and_fetch("CALL mg.procedures() YIELD *;")
            self.query_modules = [QueryModule(**module_dict) for module_dict in results]

        return (
            self.query_modules
            if starts_with is None
            else [q for q in self.query_modules if q.name.startswith(starts_with)]
        )

    def add_query_module(self, file_path: str, module_name: str) -> "Memgraph":
        """Function for adding a query module in Python written language to Memgraph.
        Example can be found in the functions below (with_kafka_stream, with_power_bi).

        The module is synced dynamically then with the database to enable higher processing
        capabilities.

        Args:
            file_name (str): path to file containing module.
            module_name (str): name of the module.

        Returns:
            Memgraph: Memgraph object.
        """
        if not os.path.isfile(file_path):
            raise GQLAlchemyFileNotFoundError(path=file_path)

        file_text = open(file_path, "r").read().replace("'", '"')
        query = f"CALL mg.create_module_file('{module_name}','{file_text}') YIELD *;"
        list(self.execute_and_fetch(query))

        return self

    def with_kafka_stream(self) -> "Memgraph":
        """Load kafka stream query module.
        Returns:
            Memgraph: Memgraph instance
        """
        file_path = "gqlalchemy/query_modules/push_streams/kafka.py"
        module_name = "kafka_stream.py"

        return self.add_query_module(file_path=file_path, module_name=module_name)

    def with_power_bi(self) -> "Memgraph":
        """Load power_bi stream query module.
        Returns:
            Memgraph: Memgraph instance
        """
        file_path = "gqlalchemy/query_modules/push_streams/power_bi.py"
        module_name = "power_bi_stream.py"

        return self.add_query_module(file_path=file_path, module_name=module_name)

    def get_storage_mode(self) -> str:
        """Returns the storage mode of the Memgraph instance."""
        result = self.execute_and_fetch("SHOW STORAGE INFO;")
        storage_mode_value = next((item["value"] for item in result if item["storage info"] == "storage_mode"), None)
        return MemgraphStorageMode(storage_mode_value).value

    def set_storage_mode(self, storage_mode: MemgraphStorageMode) -> None:
        """Sets the storage mode of the Memgraph instance."""
        self.execute(f"STORAGE MODE {storage_mode};")

    def get_transactions(self) -> List[MemgraphTransaction]:
        """Get all transactions in the database.
        Returns:
            List[MemgraphTransaction]: A list of MemgraphTransaction objects.
        """

        transactions_data = self.execute_and_fetch("SHOW TRANSACTIONS;")
        transactions = list(map(create_transaction, transactions_data))

        return transactions

    def terminate_transactions(self, transaction_ids: List[str]) -> List[MemgraphTerminatedTransaction]:
        """Terminate transactions in the database.
        Args:
            transaction_ids (List[str]): A list of transaction ids to terminate.
        Returns:
            List[MemgraphTerminatedTransaction]: A list of MemgraphTerminatedTransaction objects with info on their status.
        """

        query = (
            "TERMINATE TRANSACTIONS " + ", ".join([f"'{transaction_id}'" for transaction_id in transaction_ids]) + ";"
        )

        transactions_data = self.execute_and_fetch(query)

        terminated_transactions = list(map(create_terminated_transaction, transactions_data))

        return terminated_transactions



# Copyright (c) 2016-2022 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from gqlalchemy.query_builders.declarative_base import (  # noqa F401
    DeclarativeBase,
    Call,
    Create,
    Foreach,
    Match,
    Merge,
    Return,
    Unwind,
    With,
)
from gqlalchemy.vendors.neo4j import Neo4j


class Neo4jQueryBuilder(DeclarativeBase):
    def __init__(self, connection: Neo4j):
        super().__init__(connection)



# Copyright (c) 2016-2022 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sqlite3
import contextlib
from abc import ABC
from typing import Optional, List


class OnDiskPropertyDatabase(ABC):
    """An abstract class for implementing on-disk storage features with specific databases."""

    def save_node_property(self, node_id: int, property_name: str, property_value: str) -> None:
        """Saves a node property to an on disk database."""
        pass

    def load_node_property(self, node_id: int, property_name: str, property_value: str) -> Optional[str]:
        """Loads a node property from an on disk database."""
        pass

    def delete_node_property(self, node_id: int, property_name: str, property_value: str) -> None:
        """Deletes a node property from an on disk database."""
        pass

    def save_relationship_property(self, relationship_id: int, property_name: str, property_value: str) -> None:
        """Saves a relationship property to an on disk database."""
        pass

    def load_relationship_property(
        self, relationship_id: int, property_name: str, property_value: str
    ) -> Optional[str]:
        """Loads a relationship property from an on disk database."""
        pass

    def delete_relationship_property(self, node_id: int, property_name: str, property_value: str) -> None:
        """Deletes a node property from an on disk database."""
        pass

    def drop_database(self) -> None:
        """Deletes all entries from the on disk database."""
        pass


class SQLitePropertyDatabase(OnDiskPropertyDatabase):
    def __init__(self, database_path: str, memgraph: "Memgraph" = None):  # noqa F821
        self.database_name = database_path
        self._create_node_property_table()
        self._create_relationship_property_table()
        if memgraph is not None:
            memgraph.init_disk_storage(self)

    def execute_query(self, query: str) -> List[str]:
        """Executes an SQL query on the on disk property database.

        Args:
            query: A string representing an SQL query.

        Returns:
            A list of strings representing the results of the query.
        """
        with contextlib.closing(sqlite3.connect(self.database_name)) as conn:
            with conn:  # autocommit changes
                with contextlib.closing(conn.cursor()) as cursor:
                    cursor.execute(query)
                    return cursor.fetchall()

    def _create_node_property_table(self) -> None:
        """Creates a node property SQL table."""
        self.execute_query(
            "CREATE TABLE IF NOT EXISTS node_properties ("
            "node_id integer NOT NULL,"
            "property_name text NOT NULL,"
            "property_value text NOT NULL,"
            "PRIMARY KEY (node_id, property_name)"
            ");"
        )

    def _create_relationship_property_table(self) -> None:
        """Creates a relationship property SQL table."""
        self.execute_query(
            "CREATE TABLE IF NOT EXISTS relationship_properties ("
            "relationship_id integer NOT NULL,"
            "property_name text NOT NULL,"
            "property_value text NOT NULL,"
            "PRIMARY KEY (relationship_id, property_name)"
            ");"
        )

    def drop_database(self) -> None:
        """Deletes all properties in the database."""
        self.execute_query("DELETE FROM node_properties;")
        self.execute_query("DELETE FROM relationship_properties;")

    def save_node_property(self, node_id: int, property_name: str, property_value: str) -> None:
        """Saves a node property to an on disk database.

        Args:
            node_id: An integer representing the internal id of the node.
            property_name: A string representing the name of the property.
            property_value: A string representing the value of the property.
        """
        self.execute_query(
            "INSERT INTO node_properties (node_id, property_name, property_value) "
            f"VALUES({node_id}, '{property_name}', '{property_value}') "
            "ON CONFLICT(node_id, property_name) "
            "DO UPDATE SET property_value=excluded.property_value;"
        )

    def load_node_property(self, node_id: int, property_name: str) -> Optional[str]:
        """Loads a node property from an on disk database.

        Args:
            node_id: An integer representing the internal id of the node.
            property_name: A string representing the name of the property.

        Returns:
            An optional string representing the property value.
        """
        result = self.execute_query(
            "SELECT property_value "
            "FROM node_properties AS db "
            f"WHERE db.node_id = {node_id} "
            f"AND db.property_name = '{property_name}'"
        )

        if len(result) == 0:
            return None

        # primary key is unique
        assert len(result) == 1 and len(result[0]) == 1

        return result[0][0]

    def delete_node_property(self, node_id: int, property_name: str) -> None:
        """Deletes a node property from an on disk database.

        Args:
            node_id: An integer representing the internal id of the node.
            property_name: A string representing the name of the property.
        """
        self.execute_query(
            "DELETE "
            "FROM node_properties AS db "
            f"WHERE db.node_id = {node_id} "
            f"AND db.property_name = '{property_name}'"
        )

    def save_relationship_property(self, relationship_id: int, property_name: str, property_value: str) -> None:
        """Saves a relationship property to an on disk database.

        Args:
            relationship_id: An integer representing the internal id of the relationship.
            property_name: A string representing the name of the property.
            property_value: A string representing the value of the property.
        """
        self.execute_query(
            "INSERT INTO relationship_properties (relationship_id, property_name, property_value) "
            f"VALUES({relationship_id}, '{property_name}', '{property_value}') "
            "ON CONFLICT(relationship_id, property_name) "
            "DO UPDATE SET property_value=excluded.property_value;"
        )

    def load_relationship_property(self, relationship_id: int, property_name: str) -> Optional[str]:
        """Loads a relationship property from an on disk database.

        Args:
            relationship_id: An integer representing the internal id of the relationship.
            property_name: A string representing the name of the property.

        Returns:
            An optional string representing the property value.
        """
        result = self.execute_query(
            "SELECT property_value "
            "FROM relationship_properties AS db "
            f"WHERE db.relationship_id = {relationship_id} "
            f"AND db.property_name = '{property_name}'"
        )

        if len(result) == 0:
            return None

        # primary key is unique
        assert len(result) == 1 and len(result[0]) == 1

        return result[0][0]

    def delete_relationship_property(self, relationship_id: int, property_name: str) -> None:
        """Deletes a node property from an on disk database.

        Args:
            relationship_id: An integer representing the internal id of the relationship.
            property_name: A string representing the name of the property.
        """
        self.execute_query(
            "DELETE "
            "FROM relationship_properties AS db "
            f"WHERE db.relationship_id = {relationship_id} "
            f"AND db.property_name = '{property_name}'"
        )



# Copyright (c) 2016-2023 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import dgl
import torch

from gqlalchemy.transformations.translators.translator import Translator
from gqlalchemy.transformations.constants import DGL_ID
from gqlalchemy.utilities import to_cypher_value
from gqlalchemy.memgraph_constants import (
    MG_HOST,
    MG_PORT,
    MG_USERNAME,
    MG_PASSWORD,
    MG_ENCRYPTED,
    MG_CLIENT_NAME,
    MG_LAZY,
)


class DGLTranslator(Translator):
    """Performs conversion from cypher queries to the DGL graph representation. DGL assigns to each edge a unique integer, called the edge ID,
    based on the order in which it was added to the graph. In DGL, all the edges are directed, and an edge (u,v) indicates that the direction goes
    from node u to node v. Only features of numerical types (e.g., float, double, and int) are allowed. They can be scalars, vectors or multi-dimensional
    tensors (DGL requirement). Each node feature has a unique name and each edge feature has a unique name. The features of nodes and edges can have
    the same name. A feature is created via tensor assignment, which assigns a feature to each node/edge in the graph. The leading dimension of that
    tensor must be equal to the number of nodes/edges in the graph. You cannot assign a feature to a subset of the nodes/edges in the graph. Features of the
    same name must have the same dimensionality and data type.
    """

    def __init__(
        self,
        host: str = MG_HOST,
        port: int = MG_PORT,
        username: str = MG_USERNAME,
        password: str = MG_PASSWORD,
        encrypted: bool = MG_ENCRYPTED,
        client_name: str = MG_CLIENT_NAME,
        lazy: bool = MG_LAZY,
    ) -> None:
        super().__init__(host, port, username, password, encrypted, client_name, lazy)

    def to_cypher_queries(self, graph: Union[dgl.DGLGraph, dgl.DGLHeteroGraph]):
        """Produce cypher queries for data saved as part of the DGL graph. The method handles both homogeneous and heterogeneous graph. If the graph is homogeneous, a default DGL's labels will be used.
         _N as a node label and _E as edge label. The method converts 1D as well as multidimensional features. If there are some isolated nodes inside DGL graph, they won't get transferred. Nodes and edges
         created in Memgraph DB will, for the consistency reasons, have property `dgl_id` set to the id they have as part of the DGL graph. Note that this method doesn't insert anything inside the database,
         it just creates cypher queries. To insert queries the following code can be used:
         >>> memgraph = Memgraph()
         dgl_graph = DGLGraph(...)
         for query in DGLTranslator().to_cypher_queries(dgl_graph):
            memgraph.execute(query)

         Args:
            graph: A reference to the DGL graph.
        Returns:
            cypher queries.
        """
        queries = []
        if graph is None:
            return queries
        for etype in graph.canonical_etypes:
            source_node_label, edge_type, dest_node_label = etype
            source_nodes, dest_nodes, eids = graph.edges(etype=etype, form="all")
            # Get label and type properties
            node_src_label_properties = graph.nodes[source_node_label].data
            node_dest_label_properties = graph.nodes[dest_node_label].data
            etype_properties = graph.edges[etype].data
            for source_node_id, dest_node_id, eid in zip(source_nodes, dest_nodes, eids):
                # Handle properties
                source_node_properties, dest_node_properties, edge_properties = {}, {}, {}
                # Copy source node properties
                source_node_properties = dict(
                    map(
                        lambda pair: (pair[0], to_cypher_value(pair[1][source_node_id])),
                        node_src_label_properties.items(),
                    )
                )
                source_node_properties[DGL_ID] = int(source_node_id)
                # Copy destination node properties
                dest_node_properties = dict(
                    map(
                        lambda pair: (pair[0], to_cypher_value(pair[1][dest_node_id])),
                        node_dest_label_properties.items(),
                    )
                )
                dest_node_properties[DGL_ID] = int(dest_node_id)
                # Copy edge features
                edge_properties = dict(
                    map(lambda pair: (pair[0], to_cypher_value(pair[1][eid])), etype_properties.items())
                )
                edge_properties[DGL_ID] = int(eid)

                # Create query
                queries.append(
                    self.create_insert_query(
                        source_node_label,
                        source_node_properties,
                        edge_type,
                        edge_properties,
                        dest_node_label,
                        dest_node_properties,
                    )
                )
        return queries

    def get_instance(self) -> dgl.DGLHeteroGraph:
        """Create instance of DGL graph from all edges that are inside Memgraph. Currently, isolated nodes are ignored because they don't contribute in message passing neural networks. Only numerical features
        that are set on all nodes or all edges are transferred to the DGL instance since this is DGL's requirement. That means that any string values properties won't be transferred, as well as numerical properties
        that aren't set on all nodes. However, features of type list are transferred to the DGL and can be used as any other feature in the DGL graph. Regardless of data residing inside Memgraph database, the created
        DGL graph is a heterograph instance.
        Returns:
            DGL heterograph instance.
        """
        # Parse it into nice data structures
        src_nodes, dest_nodes, node_features, edge_features, _ = self._parse_memgraph()

        graph_data = {}
        # Iterating over src_nodes and dest_nodes should be the same
        for type_triplet in src_nodes.keys():
            graph_data[type_triplet] = (
                torch.tensor(src_nodes[type_triplet], dtype=torch.int32),
                torch.tensor(dest_nodes[type_triplet], dtype=torch.int32),
            )

        # Fail safely when graph is empty
        if len(graph_data) == 0:
            return None

        # Create heterograph
        graph = dgl.heterograph(graph_data)
        # Set node features
        for node_label, features_dict in node_features.items():
            for feature_name, features in features_dict.items():
                translated_features = Translator.validate_features(features, graph.num_nodes(node_label))
                if translated_features is None:
                    continue
                graph.nodes[node_label].data[feature_name] = translated_features

        # Set edge features
        for edge_triplet, features_dict in edge_features.items():
            for feature_name, features in features_dict.items():
                translated_features = Translator.validate_features(features, graph.num_edges(edge_triplet))
                if translated_features is None:
                    continue
                graph.edges[edge_triplet].data[feature_name] = translated_features

        return graph



# Copyright (c) 2016-2022 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, Iterator, List, Optional, Set, Tuple, Union

from gqlalchemy.exceptions import (
    GQLAlchemyExtraKeywordArguments,
    GQLAlchemyInstantiationError,
    GQLAlchemyLiteralAndExpressionMissing,
    GQLAlchemyMissingOrder,
    GQLAlchemyOperatorTypeError,
    GQLAlchemyOrderByTypeError,
    GQLAlchemyResultQueryTypeError,
    GQLAlchemyTooLargeTupleInResultQuery,
)
from gqlalchemy.graph_algorithms.integrated_algorithms import IntegratedAlgorithm
from gqlalchemy.vendors.memgraph import Memgraph
from gqlalchemy.models import Node, Relationship
from gqlalchemy.utilities import to_cypher_labels, to_cypher_properties, to_cypher_value, to_cypher_qm_arguments
from gqlalchemy.vendors.database_client import DatabaseClient


class DeclarativeBaseTypes:
    CALL = "CALL"
    CREATE = "CREATE"
    DELETE = "DELETE"
    FOREACH = "FOREACH"
    LIMIT = "LIMIT"
    LOAD_CSV = "LOAD_CSV"
    MATCH = "MATCH"
    MERGE = "MERGE"
    NODE = "NODE"
    ORDER_BY = "ORDER BY"
    RELATIONSHIP = "RELATIONSHIP"
    REMOVE = "REMOVE"
    RETURN = "RETURN"
    SET = "SET"
    SKIP = "SKIP"
    UNION = "UNION"
    UNWIND = "UNWIND"
    WHERE = "WHERE"
    WITH = "WITH"
    YIELD = "YIELD"


class MatchConstants:
    DIRECTED = "directed"
    LABELS_STR = "labels_str"
    OPTIONAL = "optional"
    PROPERTIES_STR = "properties_str"
    QUERY = "query"
    TYPE = "type"
    VARIABLE = "variable"


class Result(Enum):
    RETURN = 1
    YIELD = 2
    WITH = 3


class Where(Enum):
    WHERE = 1
    AND = 2
    OR = 3
    XOR = 4
    NOT = 5


class Operator(Enum):
    ASSIGNMENT = "="
    EQUAL = "="
    GEQ_THAN = ">="
    GREATER_THAN = ">"
    INEQUAL = "<>"
    LABEL_FILTER = ":"
    LESS_THAN = "<"
    LEQ_THAN = "<="
    NOT_EQUAL = "!="
    INCREMENT = "+="
    SIMILAR = "=~"


class Order(Enum):
    ASC = 1
    ASCENDING = 2
    DESC = 3
    DESCENDING = 4


class NoVariablesMatchedException(Exception):
    def __init__(self):
        message = "No variables have been matched in the query"
        super().__init__(message)


class InvalidMatchChainException(Exception):
    def __init__(self):
        message = "Invalid match query when linking!"
        super().__init__(message)


class PartialQuery(ABC):
    def __init__(self, type: str):
        self.type = type

    @abstractmethod
    def construct_query(self) -> str:
        pass


class MatchPartialQuery(PartialQuery):
    def __init__(self, optional: bool):
        super().__init__(DeclarativeBaseTypes.MATCH)

        self.optional = optional

    def construct_query(self) -> str:
        if self.optional:
            return " OPTIONAL MATCH "

        return " MATCH "


class MergePartialQuery(PartialQuery):
    def __init__(self):
        super().__init__(DeclarativeBaseTypes.MERGE)

    def construct_query(self) -> str:
        return " MERGE "


class CreatePartialQuery(PartialQuery):
    def __init__(self):
        super().__init__(DeclarativeBaseTypes.CREATE)

    def construct_query(self) -> str:
        return " CREATE "


class CallPartialQuery(PartialQuery):
    def __init__(self, procedure: str, arguments: Optional[Union[str, Tuple[Union[str, int, float]]]]):
        super().__init__(DeclarativeBaseTypes.CALL)

        self.procedure = procedure
        self.arguments = to_cypher_qm_arguments(arguments)

    def construct_query(self) -> str:
        return f" CALL {self.procedure}({self.arguments if self.arguments else ''}) "


class WhereConditionPartialQuery(PartialQuery):
    _LITERAL = "literal"
    _EXPRESSION = "expression"

    def __init__(self, item: str, operator: Operator, keyword: Where = Where.WHERE, is_negated: bool = False, **kwargs):
        super().__init__(type=keyword.name if not is_negated else f"{keyword.name} {Where.NOT.name}")
        self.query = self._build_where_query(item=item, operator=operator, **kwargs)

    def construct_query(self) -> str:
        """Constructs a where partial query."""
        return f" {self.type} {self.query} "

    def _build_where_query(self, item: str, operator: Operator, **kwargs) -> "DeclarativeBase":
        """Builds parts of a WHERE Cypher query divided by the boolean operators."""
        literal = kwargs.get(WhereConditionPartialQuery._LITERAL)
        value = kwargs.get(WhereConditionPartialQuery._EXPRESSION)

        operator_str = operator.value if isinstance(operator, Operator) else operator

        if operator_str not in Operator._value2member_map_:
            raise GQLAlchemyOperatorTypeError(clause=self.type)

        if value is None:
            if literal is None:
                raise GQLAlchemyLiteralAndExpressionMissing(clause=self.type)

            value = to_cypher_value(literal)
        elif literal is not None:
            raise GQLAlchemyExtraKeywordArguments(clause=self.type)

        return ("" if operator_str == Operator.LABEL_FILTER.value else " ").join(
            [
                item,
                operator_str,
                value,
            ]
        )


class WhereNotConditionPartialQuery(WhereConditionPartialQuery):
    def __init__(self, item: str, operator: Operator, keyword: Where = Where.WHERE, **kwargs):
        super().__init__(item=item, operator=operator, keyword=keyword, is_negated=True, **kwargs)


class AndWhereConditionPartialQuery(WhereConditionPartialQuery):
    def __init__(self, item: str, operator: Operator, **kwargs):
        super().__init__(item=item, operator=operator, keyword=Where.AND, **kwargs)


class AndNotWhereConditionPartialQuery(WhereNotConditionPartialQuery):
    def __init__(self, item: str, operator: Operator, **kwargs):
        super().__init__(item=item, operator=operator, keyword=Where.AND, **kwargs)


class OrWhereConditionPartialQuery(WhereConditionPartialQuery):
    def __init__(self, item: str, operator: Operator, **kwargs):
        super().__init__(item=item, operator=operator, keyword=Where.OR, **kwargs)


class OrNotWhereConditionPartialQuery(WhereNotConditionPartialQuery):
    def __init__(self, item: str, operator: Operator, **kwargs):
        super().__init__(item=item, operator=operator, keyword=Where.OR, **kwargs)


class XorWhereConditionPartialQuery(WhereConditionPartialQuery):
    def __init__(self, item: str, operator: Operator, **kwargs):
        super().__init__(item=item, operator=operator, keyword=Where.XOR, **kwargs)


class XorNotWhereConditionPartialQuery(WhereNotConditionPartialQuery):
    def __init__(self, item: str, operator: Operator, **kwargs):
        super().__init__(item=item, operator=operator, keyword=Where.XOR, **kwargs)


class NodePartialQuery(PartialQuery):
    def __init__(self, variable: Optional[str], labels: str, properties: str):
        super().__init__(DeclarativeBaseTypes.NODE)

        self._variable = "" if variable is None else variable
        self._labels = labels
        self._properties = properties

    @property
    def variable(self) -> str:
        return self._variable

    @property
    def labels(self) -> str:
        return self._labels

    @property
    def properties(self) -> str:
        return self._properties

    def construct_query(self) -> str:
        """Constructs a node partial query."""
        return f"({self.variable}{self.labels}{' ' + self.properties if self.properties else ''})"


class RelationshipPartialQuery(PartialQuery):
    def __init__(
        self,
        variable: Optional[str],
        relationship_type: str,
        algorithm: str,
        properties: str,
        directed: bool,
        from_: bool,
    ):
        super().__init__(DeclarativeBaseTypes.RELATIONSHIP)

        self.directed = directed
        self._variable = "" if variable is None else variable
        self._relationship_type = relationship_type
        self._algorithm = algorithm
        self._properties = properties
        self._from = from_

    @property
    def variable(self) -> str:
        return self._variable

    @property
    def relationship_type(self) -> str:
        return self._relationship_type

    @property
    def algorithm(self) -> str:
        return self._algorithm

    @property
    def properties(self) -> str:
        return self._properties

    def construct_query(self) -> str:
        """Constructs a relationship partial query."""
        relationship_query = f"{self.variable}{self.relationship_type}{self.algorithm}{self.properties}"

        if not self.directed:
            relationship_query = f"-[{relationship_query}]-"
            return relationship_query

        if self._from:
            relationship_query = f"<-[{relationship_query}]-"
        else:
            relationship_query = f"-[{relationship_query}]->"

        return relationship_query


class UnwindPartialQuery(PartialQuery):
    def __init__(self, list_expression: str, variable: str):
        super().__init__(DeclarativeBaseTypes.UNWIND)

        self.list_expression = list_expression
        self.variable = variable

    def construct_query(self) -> str:
        """Constructs an unwind partial query."""
        return f" UNWIND {self.list_expression} AS {self.variable} "


def dict_to_alias_statement(alias_dict: Dict[str, str]) -> str:
    """Creates a string expression of alias statements from a dictionary of
    expression, variable name dictionary.
    """
    return ", ".join(
        f"{key} AS {value}" if value != "" and key != value else f"{key}" for (key, value) in alias_dict.items()
    )


class _ResultPartialQuery(PartialQuery):
    def __init__(
        self,
        keyword: Result,
        results: Optional[
            Union[
                str,
                Tuple[str, str],
                Dict[str, str],
                List[Union[str, Tuple[str, str]]],
                Set[Union[str, Tuple[str, str]]],
            ]
        ] = None,
    ):
        if type(self) is _ResultPartialQuery:
            raise GQLAlchemyInstantiationError(class_name=type(self).__name__)

        super().__init__(type=keyword.name)

        if results is None:
            self.query = None
        elif isinstance(results, str):
            self.query = results
        elif isinstance(results, dict):
            self.query = self._return_read_dict(results)
        elif isinstance(results, tuple):
            self.query = self._return_read_tuple(results)
        elif isinstance(results, (list, set)):
            self.query = self._return_read_iterable(results)
        else:
            raise GQLAlchemyResultQueryTypeError(clause=self.type)

    def construct_query(self) -> str:
        """Creates a RETURN/YIELD/WITH statement Cypher partial query."""
        if self.query is None:
            return f" {self.type} * "

        return f" {self.type} {self.query} "

    def _return_read_iterable(
        self, iterable: Union[List[Union[str, Tuple[str, str]]], Set[Union[str, Tuple[str, str]]]]
    ):
        return ", ".join(self._return_read_item(item=item) for item in iterable)

    def _return_read_item(self, item: Union[str, Tuple]) -> str:
        if isinstance(item, str):
            return item

        if isinstance(item, tuple):
            return f"{self._return_read_tuple(item)}"

        raise GQLAlchemyResultQueryTypeError(clause=self.type)

    def _return_read_tuple(self, tuple: Tuple[str, str]) -> str:
        if len(tuple) > 2:
            raise GQLAlchemyTooLargeTupleInResultQuery(clause=self.type)

        if not isinstance(tuple[0], str) or not isinstance(tuple[1], str):
            raise GQLAlchemyResultQueryTypeError(clause=self.type)

        if tuple[0] == tuple[1] or tuple[1] == "":
            return f"{tuple[0]}"

        return f"{tuple[0]} AS {tuple[1]}"

    def _return_read_dict(self, results: Dict[str, str]):
        return f"{dict_to_alias_statement(results)}"


class WithPartialQuery(_ResultPartialQuery):
    def __init__(
        self,
        results: Optional[
            Union[
                str,
                Tuple[str, str],
                Dict[str, str],
                List[Union[str, Tuple[str, str]]],
                Set[Union[str, Tuple[str, str]]],
            ]
        ] = None,
    ):
        super().__init__(keyword=Result.WITH, results=results)


class UnionPartialQuery(PartialQuery):
    def __init__(self, include_duplicates: bool):
        super().__init__(DeclarativeBaseTypes.UNION)

        self.include_duplicates = include_duplicates

    def construct_query(self) -> str:
        """Creates a UNION statement Cypher partial query."""
        return f" UNION{f' ALL' if self.include_duplicates else ''} "


class DeletePartialQuery(PartialQuery):
    def __init__(self, variable_expressions: Union[str, List[str]], detach: bool):
        super().__init__(DeclarativeBaseTypes.DELETE)

        self._variable_expressions = variable_expressions
        self.detach = detach

    @property
    def variable_expressions(self) -> str:
        return self._variable_expressions

    def construct_query(self) -> str:
        """Creates a DELETE statement Cypher partial query."""
        return f" {'DETACH' if self.detach else ''} DELETE {', '.join(self.variable_expressions) if isinstance(self.variable_expressions, list) else self.variable_expressions} "


class RemovePartialQuery(PartialQuery):
    def __init__(self, items: Union[str, List[str]]):
        super().__init__(DeclarativeBaseTypes.REMOVE)

        self._items = items

    @property
    def items(self) -> str:
        return self._items

    def construct_query(self) -> str:
        """Creates a REMOVE statement Cypher partial query."""
        return f" REMOVE {', '.join(self.items) if isinstance(self.items, list) else self.items} "


class YieldPartialQuery(_ResultPartialQuery):
    def __init__(
        self,
        results: Optional[
            Union[
                str,
                Tuple[str, str],
                Dict[str, str],
                List[Union[str, Tuple[str, str]]],
                Set[Union[str, Tuple[str, str]]],
            ]
        ] = None,
    ):
        super().__init__(keyword=Result.YIELD, results=results)


class ReturnPartialQuery(_ResultPartialQuery):
    def __init__(
        self,
        results: Optional[
            Union[
                str,
                Tuple[str, str],
                Dict[str, str],
                List[Union[str, Tuple[str, str]]],
                Set[Union[str, Tuple[str, str]]],
            ]
        ] = None,
    ):
        super().__init__(keyword=Result.RETURN, results=results)


class OrderByPartialQuery(PartialQuery):
    def __init__(self, properties: Union[str, Tuple[str, Order], List[Union[str, Tuple[str, Order]]]]):
        super().__init__(DeclarativeBaseTypes.ORDER_BY)

        self.query = (
            self._order_by_read_list(properties)
            if isinstance(properties, list)
            else self._order_by_read_item(properties)
        )

    def construct_query(self) -> str:
        """Creates a ORDER BY statement Cypher partial query."""
        return f" {self.type} {self.query} "

    def _order_by_read_item(self, item: Union[str, Tuple[str, Order]]) -> str:
        if isinstance(item, str):
            return item
        elif isinstance(item, tuple):
            return f"{self._order_by_read_tuple(item)}"
        else:
            raise GQLAlchemyOrderByTypeError

    def _order_by_read_list(self, property: List[Union[str, Tuple[str, Order]]]):
        return ", ".join(self._order_by_read_item(item=item) for item in property)

    def _order_by_read_tuple(self, tuple: Tuple[str, Order]) -> str:
        if not isinstance(tuple[1], Order):
            raise GQLAlchemyMissingOrder

        return f"{tuple[0]} {tuple[1].name}"


class LimitPartialQuery(PartialQuery):
    def __init__(self, integer_expression: Union[str, int]):
        super().__init__(DeclarativeBaseTypes.LIMIT)

        self.integer_expression = integer_expression

    def construct_query(self) -> str:
        """Creates a LIMIT statement Cypher partial query."""
        return f" LIMIT {self.integer_expression} "


class SkipPartialQuery(PartialQuery):
    def __init__(self, integer_expression: Union[str, int]):
        super().__init__(DeclarativeBaseTypes.SKIP)

        self.integer_expression = integer_expression

    def construct_query(self) -> str:
        """Creates a SKIP statement Cypher partial query."""
        return f" SKIP {self.integer_expression} "


class AddStringPartialQuery(PartialQuery):
    def __init__(self, custom_cypher: str):
        super().__init__(DeclarativeBaseTypes.SKIP)

        self.custom_cypher = custom_cypher

    def construct_query(self) -> str:
        return f"{self.custom_cypher}"


class ForeachPartialQuery(PartialQuery):
    def __init__(self, variable: str, expression: str, update_clauses: str):
        super().__init__(DeclarativeBaseTypes.FOREACH)
        self._variable = variable
        self._expression = expression
        self._update_clauses = update_clauses

    @property
    def variable(self) -> str:
        return self._variable

    @property
    def expression(self) -> str:
        return self._expression

    @property
    def update_clauses(self) -> str:
        return self._update_clauses

    def construct_query(self) -> str:
        """Creates a FOREACH statement Cypher partial query."""
        return f" FOREACH ( {self.variable} IN {self.expression} | {self.update_clauses} ) "


class SetPartialQuery(PartialQuery):
    _LITERAL = "literal"
    _EXPRESSION = "expression"

    def __init__(self, item: str, operator: Operator, **kwargs):
        super().__init__(DeclarativeBaseTypes.SET)

        self.query = self._build_set_query(item=item, operator=operator, **kwargs)

    def construct_query(self) -> str:
        """Constructs a set partial query."""
        return f" {self.type} {self.query}"

    def _build_set_query(self, item: str, operator: Operator, **kwargs) -> "DeclarativeBase":
        """Builds parts of a SET Cypher query divided by the boolean operators."""
        literal = kwargs.get(SetPartialQuery._LITERAL)
        value = kwargs.get(SetPartialQuery._EXPRESSION)

        operator_str = operator.value if isinstance(operator, Operator) else operator

        if operator_str not in Operator._value2member_map_:
            raise GQLAlchemyOperatorTypeError(clause=self.type)

        if value is None:
            if literal is None:
                raise GQLAlchemyLiteralAndExpressionMissing(clause=self.type)

            value = to_cypher_value(literal)
        elif literal is not None:
            raise GQLAlchemyExtraKeywordArguments(clause=self.type)

        return ("" if operator_str == Operator.LABEL_FILTER.value else " ").join(
            [
                item,
                operator_str,
                value,
            ]
        )


class DeclarativeBase(ABC):
    def __init__(self, connection: Optional[DatabaseClient] = None):
        self._query: List[PartialQuery] = []
        self._connection = connection if connection is not None else Memgraph()
        self._fetch_results: bool = False

    def match(self, optional: bool = False) -> "DeclarativeBase":
        """Obtain data from the database by matching it to a given pattern.

        Args:
            optional: A bool indicating if missing parts of the pattern will be
            filled with null values.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Examples:
            Get all nodes with a certain label:

            Python: `match().node(labels='Country', variable='c').return_(results='c').execute()`
            Cypher: `MATCH (c:Country) RETURN c;`

            Get a relationship of a certain type that connects two nodes with certain label:

            Python: `match().node(labels='Town', variable='t').to(relationship_type='BELONGS_TO', variable='b').node(labels='Country', variable='c').return_(results='b').execute()`
            Cypher: `MATCH (t:Town)-[b:BELONGS_TO]->(c:Country) RETURN b;`
        """
        self._query.append(MatchPartialQuery(optional))

        return self

    def merge(self) -> "DeclarativeBase":
        """Ensure that a pattern you are looking for exists in the database.
        This means that if the pattern is not found, it will be created. In a
        way, this clause is like a combination of MATCH and CREATE.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Example:
            Merge node with properties:

            Python: `merge().node(variable='city').where(item='city.name', operator=Operator.EQUAL, literal='London').return_(results='city').execute()`
            Cypher: `MERGE (city) WHERE city.name = 'London' RETURN city;`
        """
        self._query.append(MergePartialQuery())

        return self

    def create(self) -> "DeclarativeBase":
        """Create nodes and relationships in a graph.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Example:
            Create a single node:

            Python: `create().node(labels='Person', variable='p').return_(results='p').execute()`
            Cypher: `CREATE (p:Person) RETURN p;`
        """
        self._query.append(CreatePartialQuery())

        return self

    def call(
        self, procedure: str, arguments: Optional[Union[str, Tuple[Union[str, int, float]]]] = None
    ) -> "DeclarativeBase":
        """Call a query module procedure.

        Args:
            procedure: A string representing the name of the procedure in the
              format `query_module.procedure`.
            arguments: A string representing the arguments of the procedure in
              text format.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Examples:
            Call procedure with no arguments:

            Python: `call('pagerank.get').yield_().return_().execute()`
            Cypher: `CALL pagerank.get() YIELD * RETURN *;`

            Call procedure with arguments:

            Python: `call('json_util.load_from_url', "'https://some-url.com'").yield_('objects').return_(results='objects').execute()
            Cypher: `CALL json_util.load_from_url(https://some-url.com) YIELD objects RETURN objects;`
        """
        self._query.append(CallPartialQuery(procedure, arguments))

        return self

    def node(
        self,
        labels: Union[str, List[str], None] = "",
        variable: Optional[str] = None,
        node: Optional["Node"] = None,
        **kwargs,
    ) -> "DeclarativeBase":
        """Add a node pattern to the query.

        Args:
            labels: A string or list of strings representing the labels of the
              node.
            variable: A string representing the name of the variable for storing
              results of the node pattern.
            node: A `Node` object to construct the pattern from.
            **kwargs: Arguments representing the properties of the node.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Example:
            Create a node and return it:

            Python: `create().node(labels='Person', variable='n', first_name='Kate').return_(results='n').execute()`
            Cypher: `CREATE (n:Person {first_name: 'Kate'}) RETURN n;`
        """
        if not self._is_linking_valid_with_query(DeclarativeBaseTypes.NODE):
            raise InvalidMatchChainException()

        if node is None:
            labels_str = to_cypher_labels(labels)
            properties_str = to_cypher_properties(kwargs)
        else:
            labels_str = to_cypher_labels(node._labels)
            properties_str = to_cypher_properties(node._properties)

        self._query.append(NodePartialQuery(variable, labels_str, properties_str))

        return self

    def to(
        self,
        relationship_type: Optional[str] = "",
        directed: Optional[bool] = True,
        variable: Optional[str] = None,
        relationship: Optional["Relationship"] = None,
        algorithm: Optional[IntegratedAlgorithm] = None,
        **kwargs,
    ) -> "DeclarativeBase":
        """Add a relationship pattern to the query.

        Args:
            relationship_type: A string representing the type of the relationship.
            directed: A bool indicating if the relationship is directed.
            variable: A string representing the name of the variable for storing
              results of the relationship pattern.
            relationship: A `Relationship` object to construct the pattern from.
            algorithm: algorithm object to use over graph data.
            **kwargs: Arguments representing the properties of the relationship.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Example:
            Match and return a relationship:

            Python: `match().node(labels='Town', variable='t').to(relationship_type='BELONGS_TO', variable='b').node(labels='Country', variable='c').return_(results='b').execute()`
            Cypher: `MATCH (t:Town)-[b:BELONGS_TO]->(c:Country) RETURN b;`
        """
        if not self._is_linking_valid_with_query(DeclarativeBaseTypes.RELATIONSHIP):
            raise InvalidMatchChainException()

        if relationship is None:
            type_str = to_cypher_labels(relationship_type)
            properties_str = to_cypher_properties(kwargs)
        else:
            type_str = to_cypher_labels(relationship._type)
            properties_str = to_cypher_properties(relationship._properties)

        self._query.append(
            RelationshipPartialQuery(
                variable=variable,
                relationship_type=type_str,
                algorithm="" if algorithm is None else str(algorithm),
                properties=properties_str,
                directed=bool(directed),
                from_=False,
            )
        )

        return self

    def from_(
        self,
        relationship_type: Optional[str] = "",
        directed: Optional[bool] = True,
        variable: Optional[str] = None,
        relationship: Optional["Relationship"] = None,
        algorithm: Optional[IntegratedAlgorithm] = None,
        **kwargs,
    ) -> "Match":
        """Add a relationship pattern to the query.

        Args:
            relationship_type: A string representing the type of the relationship.
            directed: A bool indicating if the relationship is directed.
            variable: A string representing the name of the variable for storing
              results of the relationship pattern.
            relationship: A `Relationship` object to construct the pattern from.
            **kwargs: Arguments representing the properties of the relationship.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Example:
            Match and return a relationship:

            Python: `match().node(labels='Country', variable='c').from_(relationship_type='BELONGS_TO', variable='b').node(labels='Town', variable='t').return_(results='b').execute()`
            Cypher: `MATCH (c:Country)<-[b:BELONGS_TO]-(t:Town) RETURN b;`
        """
        if not self._is_linking_valid_with_query(DeclarativeBaseTypes.RELATIONSHIP):
            raise InvalidMatchChainException()

        if relationship is None:
            type_str = to_cypher_labels(relationship_type)
            properties_str = to_cypher_properties(kwargs)
        else:
            type_str = to_cypher_labels(relationship._type)
            properties_str = to_cypher_properties(relationship._properties)

        self._query.append(
            RelationshipPartialQuery(
                variable=variable,
                relationship_type=type_str,
                algorithm="" if algorithm is None else str(algorithm),
                properties=properties_str,
                directed=bool(directed),
                from_=True,
            )
        )

        return self

    def where(self, item: str, operator: Operator, **kwargs) -> "DeclarativeBase":
        """Creates a WHERE statement Cypher partial query.

        Args:
            item: A string representing variable or property.
            operator: A string representing the operator.

        Kwargs:
            literal: A value that will be converted to Cypher value, such as int, float, string, etc.
            expression: A node label or property that won't be converted to Cypher value (no additional quotes will be added).

        Raises:
            GQLAlchemyLiteralAndExpressionMissingInWhere: Raises an error when neither literal nor expression keyword arguments were provided.
            GQLAlchemyExtraKeywordArgumentsInWhere: Raises an error when both literal and expression keyword arguments were provided.

        Returns:
            self: A partial Cypher query built from the given parameters.

        Examples:
            Filtering query results by the equality of `name` properties of two connected nodes.

            Python: `match().node(variable='n').to().node(variable='m').where(item='n.name', operator=Operator.EQUAL, expression='m.name').return_()`
            Cypher: `MATCH (n)-[]->(m) WHERE n.name = m.name RETURN *;`

            Filtering query results by the node label.

            Python: `match().node(variable='n').where(item='n', operator=Operator.LABEL_FILTER, expression='User').return_()`
            Cypher: `MATCH (n) WHERE n:User RETURN *;`

            Filtering query results by the comparison of node property and literal.

            Python: `match().node(variable='n').where(item='n.age', operator=Operator.GREATER_THAN, literal=18).return_()`
            Cypher: `MATCH (n) WHERE n.age > 18 RETURN *;`
        """
        # WHERE item operator (literal | expression)
        # item: variable | property
        # expression: label | property
        self._query.append(WhereConditionPartialQuery(item=item, operator=operator, **kwargs))

        return self

    def where_not(self, item: str, operator: Operator, **kwargs) -> "DeclarativeBase":
        """Creates a WHERE NOT statement Cypher partial query.

        Args:
            item: A string representing variable or property.
            operator: A string representing the operator.

        Kwargs:
            literal: A value that will be converted to Cypher value, such as int, float, string, etc.
            expression: A node label or property that won't be converted to Cypher value (no additional quotes will be added).

        Raises:
            GQLAlchemyLiteralAndExpressionMissingInWhere: Raises an error when neither literal nor expression keyword arguments were provided.
            GQLAlchemyExtraKeywordArgumentsInWhere: Raises an error when both literal and expression keyword arguments were provided.

        Returns:
            self: A partial Cypher query built from the given parameters.

        Examples:
            Filtering query results by the equality of `name` properties of two connected nodes.

            Python: `match().node(variable='n').to().node(variable='m').where_not(item='n.name', operator='=', expression='m.name').return_()`
            Cypher: `MATCH (n)-[]->(m) WHERE NOT n.name = m.name RETURN *;`
        """
        self._query.append(WhereNotConditionPartialQuery(item=item, operator=operator, **kwargs))

        return self

    def and_where(self, item: str, operator: Operator, **kwargs) -> "DeclarativeBase":
        """Creates an AND statement as a part of WHERE Cypher partial query.

        Args:
            item: A string representing variable or property.
            operator: A string representing the operator.

        Kwargs:
            literal: A value that will be converted to Cypher value, such as int, float, string, etc.
            expression: A node label or property that won't be converted to Cypher value (no additional quotes will be added).

        Returns:
            self: A partial Cypher query built from the given parameters.

        Examples:
            Filtering query results by node label or the comparison of node property and literal.

            Python: `match().node(variable='n').where(item='n', operator=Operator.LABEL_FILTER, expression='User').and_where(item='n.age', operator=Operator.GREATER_THAN, literal=18).return_()`
            Cypher: `MATCH (n) WHERE n:User AND n.age > 18 RETURN *;`
        """
        self._query.append(AndWhereConditionPartialQuery(item=item, operator=operator, **kwargs))

        return self

    def and_not_where(self, item: str, operator: Operator, **kwargs) -> "DeclarativeBase":
        """Creates an AND NOT statement as a part of WHERE Cypher partial query.

        Args:
            item: A string representing variable or property.
            operator: A string representing the operator.

        Kwargs:
            literal: A value that will be converted to Cypher value, such as int, float, string, etc.
            expression: A node label or property that won't be converted to Cypher value (no additional quotes will be added).

        Returns:
            self: A partial Cypher query built from the given parameters.

        Examples:
            Filtering query results by node label or the comparison of node property and literal.

            Python: `match().node(variable='n').where(item='n', operator=Operator.LABEL_FILTER, expression='User').and_not_where(item='n.age', operator=Operator.GREATER_THAN, literal=18).return_()`
            Cypher: `MATCH (n) WHERE n:User AND NOT n.age > 18 RETURN *;`
        """
        self._query.append(AndNotWhereConditionPartialQuery(item=item, operator=operator, **kwargs))

        return self

    def or_where(self, item: str, operator: Operator, **kwargs) -> "DeclarativeBase":
        """Creates an OR statement as a part of WHERE Cypher partial query.

        Args:
            item: A string representing variable or property.
            operator: A string representing the operator.

        Kwargs:
            literal: A value that will be converted to Cypher value, such as int, float, string, etc.
            expression: A node label or property that won't be converted to Cypher value (no additional quotes will be added).

        Returns:
            self: A partial Cypher query built from the given parameters.

        Examples:
            Filtering query results by node label or the comparison of node property and literal.

            Python: `match().node(variable='n').where(item='n', operator=Operator.LABEL_FILTER, expression='User').or_where(item='n.age', operator=Operator.GREATER_THAN, literal=18).return_()`
            Cypher: `MATCH (n) WHERE n:User OR n.age > 18 RETURN *;`
        """
        self._query.append(OrWhereConditionPartialQuery(item=item, operator=operator, **kwargs))

        return self

    def or_not_where(self, item: str, operator: Operator, **kwargs) -> "DeclarativeBase":
        """Creates an OR NOT statement as a part of WHERE Cypher partial query.

        Args:
            item: A string representing variable or property.
            operator: A string representing the operator.

        Kwargs:
            literal: A value that will be converted to Cypher value, such as int, float, string, etc.
            expression: A node label or property that won't be converted to Cypher value (no additional quotes will be added).

        Returns:
            self: A partial Cypher query built from the given parameters.

        Examples:
            Filtering query results by node label or the comparison of node property and literal.

            Python: `match().node(variable='n').where(item='n', operator=Operator.LABEL_FILTER, expression='User').or_not_where(item='n.age', operator=Operator.GREATER_THAN, literal=18).return_()`
            Cypher: `MATCH (n) WHERE n:User OR NOT n.age > 18 RETURN *;`
        """
        self._query.append(OrNotWhereConditionPartialQuery(item=item, operator=operator, **kwargs))

        return self

    def xor_where(self, item: str, operator: Operator, **kwargs) -> "DeclarativeBase":
        """Creates an XOR statement as a part of WHERE Cypher partial query.

        Args:
            item: A string representing variable or property.
            operator: A string representing the operator.

        Kwargs:
            literal: A value that will be converted to Cypher value, such as int, float, string, etc.
            expression: A node label or property that won't be converted to Cypher value (no additional quotes will be added).

        Returns:
            self: A partial Cypher query built from the given parameters.

        Examples:
            Filtering query results by node label or the comparison of node property and literal.

            Python: `match().node(variable='n').where(item='n', operator=Operator.LABEL_FILTER, expression='User').xor_where(item='n.age', operator=Operator.GREATER_THAN, literal=18).return_()`
            Cypher: `MATCH (n) WHERE n:User XOR n.age > 18 RETURN *;`
        """
        self._query.append(XorWhereConditionPartialQuery(item=item, operator=operator, **kwargs))

        return self

    def xor_not_where(self, item: str, operator: Operator, **kwargs) -> "DeclarativeBase":
        """Creates an XOR NOT statement as a part of WHERE Cypher partial query.

        Args:
            item: A string representing variable or property.
            operator: A string representing the operator.

        Kwargs:
            literal: A value that will be converted to Cypher value, such as int, float, string, etc.
            expression: A node label or property that won't be converted to Cypher value (no additional quotes will be added).

        Returns:
            self: A partial Cypher query built from the given parameters.

        Examples:
            Filtering query results by node label or the comparison of node property and literal.

            Python: `match().node(variable='n').where(item='n', operator=Operator.LABEL_FILTER, expression='User').xor_not_where(item='n.age', operator=Operator.GREATER_THAN, literal=18).return_()`
            Cypher: `MATCH (n) WHERE n:User XOR NOT n.age > 18 RETURN *;`
        """
        self._query.append(XorNotWhereConditionPartialQuery(item=item, operator=operator, **kwargs))

        return self

    def unwind(self, list_expression: str, variable: str) -> "DeclarativeBase":
        """Unwind a list of values as individual rows.

        Args:
            list_expression: A list of strings representing the list of values.
            variable: A string representing the variable name for unwinding results.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Example:
            Python: `unwind(list_expression="[1, 2, 3, null]", variable="x").return_(results=["x", ("'val'", "y")]).execute()`
            Cypher: `UNWIND [1, 2, 3, null] AS x RETURN x, 'val' AS y;`
        """
        self._query.append(UnwindPartialQuery(list_expression, variable))

        return self

    def with_(
        self,
        results: Optional[
            Union[
                str,
                Tuple[str, str],
                Dict[str, str],
                List[Union[str, Tuple[str, str]]],
                Set[Union[str, Tuple[str, str]]],
            ]
        ] = None,
    ) -> "DeclarativeBase":
        """Chain together parts of a query, piping the results from one to be
        used as starting points or criteria in the next.

        Args:
            results: A dictionary mapping variables in the first query with
            aliases in the second query.

        Raises:
            GQLAlchemyResultQueryTypeError: Raises an error when the provided argument is of wrong type.
            GQLAlchemyTooLargeTupleInResultQuery: Raises an error when the given tuple has length larger than 2.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Example:
            Pipe the result from first part of the query for the further use:

            Python: `match().node(variable='n').with('n').execute()`
            Cypher: `MATCH (n) WITH n;
        """
        self._query.append(WithPartialQuery(results))

        return self

    def union(self, include_duplicates: Optional[bool] = True) -> "DeclarativeBase":
        """Combine the result of multiple queries.

        Args:
            include_duplicates: A bool indicating if duplicates should be
              included.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Examples:
            Combine queries and retain duplicates:

            Python: `match().node(variable="c", labels="Country").return_(results=("c.name", "columnName")).union().match().node(variable="p", labels="Person").return_(results=("p.name", "columnName")).execute()`
            Cypher: `MATCH (c:Country) RETURN c.name AS columnName UNION ALL MATCH (p:Person) RETURN p.name AS columnName;`

            Combine queries and remove duplicates:

            Python: `match().node(variable="c", labels="Country").return_(results=("c.name", "columnName")).union(include_duplicates=False).match().node(variable="p", labels="Person").return_(results=("p.name", "columnName")).execute()`
            Cypher: `MATCH (c:Country) RETURN c.name AS columnName UNION MATCH (p:Person) RETURN p.name AS columnName;`
        """
        self._query.append(UnionPartialQuery(include_duplicates))

        return self

    def delete(self, variable_expressions: Union[str, List[str]], detach: Optional[bool] = False) -> "DeclarativeBase":
        """Delete nodes and relationships from the database.

        Args:
            variable_expressions: A string or list of strings indicating which node(s)
              and/or relationship(s) should be removed.
            detach: A bool indicating if relationships should be deleted along
              with a node.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Example:
            Delete a node:

            Python: `match().node(labels='Node1', variable='n1').delete(variable_expressions='n1').execute()`
            Cypher: `MATCH (n1:Node1) DELETE n1;`
        """
        self._query.append(DeletePartialQuery(variable_expressions, detach))

        return self

    def remove(self, items: Union[str, List[str]]) -> "DeclarativeBase":
        """Remove labels and properties from nodes and relationships.

        Args:
            items: A string or list of strings indicating which label(s) and/or properties
              should be removed.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Example:
            Remove a property from a node:

            Python: `match().node(labels='Country', variable='n', name='United Kingdom').remove(items='n.name').return_(results='n').execute()`
            Cypher: `MATCH (n:Country {name: 'United Kingdom'}) REMOVE n.name RETURN n;`
        """
        self._query.append(RemovePartialQuery(items))

        return self

    def yield_(
        self,
        results: Optional[
            Union[
                str,
                Tuple[str, str],
                Dict[str, str],
                List[Union[str, Tuple[str, str]]],
                Set[Union[str, Tuple[str, str]]],
            ]
        ] = None,
    ) -> "DeclarativeBase":
        """Yield data from the query.

        Args:
            results: A dictionary mapping items that are returned with alias names.

        Raises:
            GQLAlchemyResultQueryTypeError: Raises an error when the provided argument is of wrong type.
            GQLAlchemyTooLargeTupleInResultQuery: Raises an error when the given tuple has length larger than 2.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Examples:
            Yield all data from a query:

            Python: `call(procedure='pagerank.get').yield_().return_().execute()`
            Cypher: `CALL pagerank.get() YIELD * RETURN *;`

            Yield some data from a query:

            Python: `.call(procedure='pagerank.get').yield_(results=['node', 'rank']).return_(results=['node','rank']).execute()`
            Cypher: `CALL pagerank.get() YIELD node, rank RETURN node, rank;`
        """
        self._query.append(YieldPartialQuery(results))

        return self

    def return_(
        self,
        results: Optional[
            Union[
                str,
                Tuple[str, str],
                Dict[str, str],
                List[Union[str, Tuple[str, str]]],
                Set[Union[str, Tuple[str, str]]],
            ]
        ] = None,
    ) -> "DeclarativeBase":
        """Return data from the query.

        Args:
            results: An optional string, tuple or iterable of strings and tuples for alias names.

        Raises:
            GQLAlchemyResultQueryTypeError: Raises an error when the provided argument is of wrong type.
            GQLAlchemyTooLargeTupleInResultQuery: Raises an error when the given tuple has length larger than 2.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Examples:
            Return all variables from a query:

            Python: `match().node(labels='Person', variable='p').return_().execute()`
            Cypher: `MATCH (p:Person) RETURN *;`

            Return specific variables from a query:

            Python: `match().node(labels='Person', variable='p1').to().node(labels='Person', variable='p2').return_(results=[('p1','first'), 'p2']).execute()`
            Cypher: `MATCH (p1:Person)-[]->(p2:Person) RETURN p1 AS first, p2;`
        """
        self._query.append(ReturnPartialQuery(results=results))
        self._fetch_results = True

        return self

    def order_by(
        self, properties: Union[str, Tuple[str, Order], List[Union[str, Tuple[str, Order]]]]
    ) -> "DeclarativeBase":
        """Creates an ORDER BY statement Cypher partial query.

        Args:
            properties: Properties and order (DESC/DESCENDING/ASC/ASCENDING) by which the query results will be ordered.

        Raises:
            GQLAlchemyOrderByTypeError: Raises an error when the given ordering is of the wrong type.
            GQLAlchemyMissingOrder: Raises an error when the given property is neither string nor tuple.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Examples:
            Ordering query results by the property `n.name` in ascending order
            and by the property `n.last_name` in descending order:

            Python: `match().node(variable='n').return_().order_by(properties=['n.name', ('n.last_name', Order.DESC)]).execute()`
            Cypher: `MATCH (n) RETURN * ORDER BY n.name, n.last_name DESC;`
        """
        self._query.append(OrderByPartialQuery(properties=properties))

        return self

    def limit(self, integer_expression: Union[str, int]) -> "DeclarativeBase":
        """Limit the number of records when returning results.

        Args:
            integer_expression: An integer indicating how many records to limit
              the results to.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Example:
            Limit the number of returned results:

            Python: `match().node(labels='Person', variable='p').return_().limit(integer_expression='10').execute()`
            Cypher: `MATCH (p:Person) RETURN * LIMIT 10;`
        """
        self._query.append(LimitPartialQuery(integer_expression))

        return self

    def skip(self, integer_expression: Union[str, int]) -> "DeclarativeBase":
        """Skip a number of records when returning results.

        Args:
            integer_expression: An integer indicating how many records to skip
              in the results.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Example:
            Skip the first result:

            Python: `match().node(variable='n').return_(results='n').skip(integer_expression='1').execute()`
            Cypher: `MATCH (n) RETURN n SKIP 1;`
        """
        self._query.append(SkipPartialQuery(integer_expression))

        return self

    def add_custom_cypher(self, custom_cypher: str) -> "DeclarativeBase":
        """Inject custom Cypher code into the query.

        Args:
            custom_cypher: A string representing the Cypher code to be injected
              into the query.

        Returns:
            A `DeclarativeBase` instance for constructing queries.
        """
        self._query.append(AddStringPartialQuery(custom_cypher))
        if " RETURN " in custom_cypher:
            self._fetch_results = True

        return self

    def get_single(self, retrieve: str) -> Any:
        """Returns a single result with a `retrieve` variable name.

        Args:
            retrieve: A string representing the results variable to be returned.

        Returns:
            An iterator of dictionaries containing the results of the query.
        """
        query = self._construct_query()

        result = next(self._connection.execute_and_fetch(query), None)

        if result:
            return result[retrieve]
        return result

    def foreach(
        self, variable: str, expression: str, update_clause: Union[str, List[str], Set[str]]
    ) -> "DeclarativeBase":
        """Iterate over a list of elements and for every iteration run every update clause.

        Args:
            variable: The variable name that stores each element.
            expression: Any expression that results in a list.
            update_clauses: One or more Cypher update clauses:
                SET, REMOVE, CREATE, MERGE, DELETE, FOREACH.

        Returns:
            A `DeclarativeBase` instance for constructing queries.

        Example:
            For each number in a list, create a node:

            Python: `update_clause = QueryBuilder().create().node(variable="n", id=PropertyVariable("i"))`
                    `query_builder = QueryBuilder().foreach("i", "[1, 2, 3]", update_clause.construct_query())`
            Cypher: `FOREACH ( i IN [1, 2, 3] | CREATE (n {id: i}) )`
        """
        if isinstance(update_clause, list):
            update_clause = " ".join(update_clause)

        self._query.append(ForeachPartialQuery(variable, expression, update_clause))

        return self

    def set_(self, item: str, operator: Operator, **kwargs):
        """Creates a SET statement Cypher partial query.

        Args:
            item: A string representing variable or property.
            operator: An assignment, increment or label filter operator.

        Kwargs:
            literal: A value that will be converted to Cypher value, such as int, float, string, etc.
            expression: A node label or property that won't be converted to Cypher value (no additional quotes will be added).

        Raises:
            GQLAlchemyLiteralAndExpressionMissingInWhere: Raises an error when neither literal nor expression keyword arguments were provided.
            GQLAlchemyExtraKeywordArgumentsInWhere: Raises an error when both literal and expression keyword arguments were provided.

        Returns:
            self: A partial Cypher query built from the given parameters.

        Examples:
            Set or update a property.

            Python: `match().node(variable='n').where(item='n.name', operator=Operator.EQUAL, literal='Germany').set_(item='n.population', operator=Operator.ASSIGNMENT, literal=83000001).return_().execute()`
            Cypher: `MATCH (n) WHERE n.name = 'Germany' SET n.population = 83000001 RETURN *;`

            Set or update multiple properties.

            Python: `match().node(variable='n').where(item='n.name', operator=Operator.EQUAL, literal='Germany').set_(item='n.population', operator=Operator.ASSIGNMENT, literal=83000001).set_(item='n.capital', operator=Operator.ASSIGNMENT, literal='Berlin').return_().execute()`
            Cypher: `MATCH (n) WHERE n.name = 'Germany' SET n.population = 83000001 SET n.capital = 'Berlin' RETURN *;`

            Set node label.

            Python: `match().node(variable='n').where(item='n.name', operator=Operator.EQUAL, literal='Germany').set_(item='n', operator=Operator.LABEL_FILTER, expression='Land').return_().execute()`
            Cypher: `MATCH (n) WHERE n.name = 'Germany' SET n:Land RETURN *;`

            Replace all properties using map.

            Python: `match().node(variable='c', labels='Country').where(item='c.name', operator=Operator.EQUAL, literal='Germany').set_(item='c', operator=Operator.ASSIGNMENT, literal={'name': 'Germany', 'population': '85000000'}).return_().execute()`
            Cypher: `MATCH (c:Country) WHERE c.name = 'Germany' SET c = {name: 'Germany', population: '85000000'} RETURN *;`

            Update all properties using map.

            Python: `match().node(variable='c', labels='Country').where(item='c.name', operator=Operator.EQUAL, literal='Germany').set_(item='c', operator=Operator.INCREMENT, literal={'name': 'Germany', 'population': '85000000'}).return_().execute()`
            Cypher: `MATCH (c:Country) WHERE c.name = 'Germany' SET c += {name: 'Germany', population: '85000000'} RETURN *;`

        """
        self._query.append(SetPartialQuery(item=item, operator=operator, **kwargs))

        return self

    def execute(self) -> Iterator[Dict[str, Any]]:
        """Executes the Cypher query and returns the results.

        Returns:
            An iterator of dictionaries containing the results of the query.
        """
        query = self._construct_query()
        if self._fetch_results:
            return self._connection.execute_and_fetch(query)
        else:
            return self._connection.execute(query)

    def _construct_query(self) -> str:
        """Constructs the (partial) Cypher query so it can be executed."""
        query = [""]

        # if not self._any_variables_matched():
        #    raise NoVariablesMatchedException()

        for partial_query in self._query:
            query.append(partial_query.construct_query())

        joined_query = "".join(query)
        joined_query = re.sub("\\s\\s+", " ", joined_query)
        return joined_query

    def construct_query(self) -> str:
        return self._construct_query()

    def _any_variables_matched(self) -> bool:
        """Checks if any variables are present in the result."""
        return any(
            q.type in [DeclarativeBaseTypes.RELATIONSHIP, DeclarativeBaseTypes.NODE] and q.variable not in [None, ""]
            for q in self._query
        )

    def _is_linking_valid_with_query(self, match_type: str):
        """Checks if linking functions match the current query."""
        return len(self._query) == 0 or self._query[-1].type != match_type


class Create(DeclarativeBase):
    def __init__(self, connection: Optional[DatabaseClient] = None):
        super().__init__(connection)
        self._query.append(CreatePartialQuery())


class Match(DeclarativeBase):
    def __init__(self, optional: bool = False, connection: Optional[DatabaseClient] = None):
        super().__init__(connection)
        self._query.append(MatchPartialQuery(optional))


class Merge(DeclarativeBase):
    def __init__(self, connection: Optional[DatabaseClient] = None):
        super().__init__(connection)
        self._query.append(MergePartialQuery())


class Call(DeclarativeBase):
    def __init__(self, procedure: str, arguments: Optional[str] = None, connection: Optional[DatabaseClient] = None):
        super().__init__(connection)
        self._query.append(CallPartialQuery(procedure, arguments))


class Unwind(DeclarativeBase):
    def __init__(self, list_expression: str, variable: str, connection: Optional[DatabaseClient] = None):
        super().__init__(connection)
        self._query.append(UnwindPartialQuery(list_expression, variable))


class With(DeclarativeBase):
    def __init__(self, results: Optional[Dict[str, str]] = {}, connection: Optional[DatabaseClient] = None):
        super().__init__(connection)
        self._query.append(WithPartialQuery(results))


class Foreach(DeclarativeBase):
    def __init__(
        self,
        variable: str,
        expression: str,
        update_clauses: Union[str, List[str], Set[str]],
        connection: Optional[DatabaseClient] = None,
    ):
        super().__init__(connection)
        self._query.append(ForeachPartialQuery(variable, expression, update_clauses))


class Return(DeclarativeBase):
    def __init__(self, results: Optional[Dict[str, str]] = {}, connection: Optional[DatabaseClient] = None):
        super().__init__(connection)
        self._query.append(ReturnPartialQuery(results))
        self._fetch_results = True



# Copyright (c) 2016-2022 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import socket
import subprocess
import time
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, Union

try:
    import docker
except ModuleNotFoundError:
    docker = None
import psutil

from gqlalchemy.exceptions import (
    GQLAlchemyWaitForConnectionError,
    GQLAlchemyWaitForDockerError,
    GQLAlchemyWaitForPortError,
    raise_if_not_imported,
)
from gqlalchemy.vendors.memgraph import Memgraph


MEMGRAPH_DEFAULT_BINARY_PATH = "/usr/lib/memgraph/memgraph"
MEMGRAPH_DEFAULT_PORT = 7687
MEMGRAPH_CONFIG_BOLT_PORT = "--bolt-port"
MEMGRAPH_CONFIG_BOLT_ADDRESS = "--bolt-address"
DOCKER_IMAGE_TAG_LATEST = "latest"
LOOPBACK_ADDRESS = "127.0.0.1"
WILDCARD_ADDRESS = "0.0.0.0"


class DockerImage(Enum):
    MEMGRAPH = "memgraph/memgraph"
    MAGE = "memgraph/memgraph-mage"


class DockerContainerStatus(Enum):
    EXITED = "exited"
    PAUSED = "paused"
    RESTARTING = "restarting"
    RUNNING = "running"


def wait_for_port(
    host: str = LOOPBACK_ADDRESS,
    port: int = MEMGRAPH_DEFAULT_PORT,
    delay: float = 0.01,
    timeout: float = 5.0,
    backoff: int = 2,
) -> None:
    """Wait for a TCP port to become available.

    Args:
        host: A string representing the IP address that is being checked.
        port: A string representing the port that is being checked.
        delay: A float that defines how long to wait between retries.
        timeout: A float that defines how long to wait for the port.
        backoff: An integer used for multiplying the delay.

    Raises:
      TimeoutError: Raises an error when the host and port are not accepting
        connections after the timeout period has passed.
    """
    start_time = time.perf_counter()
    while True:
        try:
            with socket.create_connection((host, port), timeout=timeout):
                break
        except OSError as ex:
            time.sleep(delay)
            if time.perf_counter() - start_time >= timeout:
                raise GQLAlchemyWaitForPortError(port=port, host=host) from ex

        delay *= backoff


def wait_for_docker_container(container, delay: float = 0.01, timeout: float = 5.0, backoff: int = 2) -> None:
    """Wait for a Docker container to enter the status `running`.

    Args:
        container: The Docker container to wait for.
        delay: A float that defines how long to wait between retries.
        timeout: A float that defines how long to wait for the status.
        backoff: An integer used for multiplying the delay.

    Raises:
      TimeoutError: Raises an error when the container isn't running after the
        timeout period has passed.
    """
    start_time = time.perf_counter()
    while container.status != DockerContainerStatus.RUNNING.value:
        container.reload()
        time.sleep(delay)

        if time.perf_counter() - start_time >= timeout:
            raise GQLAlchemyWaitForDockerError

        delay *= backoff


class MemgraphInstance(ABC):
    def __init__(
        self,
        host: str = WILDCARD_ADDRESS,
        port: int = MEMGRAPH_DEFAULT_PORT,
        config: Dict[str, Union[str, int, bool]] = dict(),
    ) -> None:
        self.host = host
        self.port = port
        self.config = config
        self.proc_mg = None
        self.config[MEMGRAPH_CONFIG_BOLT_PORT] = self.port
        self.config[MEMGRAPH_CONFIG_BOLT_ADDRESS] = self.host
        self._memgraph = None

    @property
    def memgraph(self) -> Memgraph:
        if self._memgraph is None:
            self._memgraph = Memgraph(self.host, self.port)

        return self._memgraph

    def set_config(self, config: Dict[str, Union[str, int, bool]]) -> None:
        self.config.update(config)

    def connect(self) -> "Memgraph":
        if not self.is_running():
            raise GQLAlchemyWaitForConnectionError

        return self.memgraph

    def start_and_connect(self, restart: bool = False) -> "Memgraph":
        """Start the Memgraph instance and return the
          connection object.

        Attributes:
            restart: A bool indicating if the instance should be
              restarted if it's already running.
        """
        self.start(restart=restart)

        return self.connect()

    def start(self, restart: bool = False) -> None:
        """Start the Memgraph instance.

        Attributes:
            restart: A bool indicating if the instance should be
              restarted if it's already running.
        """
        if not restart and self.is_running():
            return

        self.stop()
        self._start_instance()

    def stop(self) -> Any:
        """Stop the Memgraph instance."""
        if not self.is_running():
            return

        self._stop_instance()

    @abstractmethod
    def is_running(self) -> bool:
        pass

    @abstractmethod
    def _start_instance(self) -> None:
        pass

    @abstractmethod
    def _stop_instance(self) -> Any:
        pass


class MemgraphInstanceBinary(MemgraphInstance):
    """A class for managing Memgraph instances started from binary files on Unix
    systems.

    Attributes:
      binary_path: A string representing the path to a Memgraph binary
        file.
      user: A string representing the user that should start the Memgraph
        process.
    """

    def __init__(self, binary_path: str = MEMGRAPH_DEFAULT_BINARY_PATH, user: str = "", **data) -> None:
        super().__init__(**data)
        self.binary_path = binary_path
        self.user = user

    def _start_instance(self) -> None:
        args_mg = f"{self.binary_path } " + (" ").join([f"{k}={v}" for k, v in self.config.items()])
        if self.user != "":
            args_mg = f"runuser -l {self.user} -c '{args_mg}'"

        self.proc_mg = subprocess.Popen(args_mg, shell=True)
        wait_for_port(self.host, self.port)

    def _stop_instance(self) -> None:
        procs = set()
        process = psutil.Process(self.proc_mg.pid)
        procs.add(process)
        for proc in process.children(recursive=True):
            procs.add(proc)
            proc.kill()

        process.kill()
        psutil.wait_procs(procs)

    def is_running(self) -> bool:
        """Check if the Memgraph instance is still running."""
        return self.proc_mg is not None and self.proc_mg.poll() is None


class MemgraphInstanceDocker(MemgraphInstance):
    """A class for managing Memgraph instances started in Docker containers.

    Attributes:
      docker_image: An enum representing the Docker image. Values:
        `DockerImage.MEMGRAPH` and `DockerImage.MAGE`.
      docker_image_tag: A string representing the tag of the Docker image.
    """

    def __init__(
        self, docker_image: DockerImage = DockerImage.MEMGRAPH, docker_image_tag: str = DOCKER_IMAGE_TAG_LATEST, **data
    ) -> None:
        raise_if_not_imported(dependency=docker, dependency_name="docker")

        super().__init__(**data)
        self.docker_image = docker_image
        self.docker_image_tag = docker_image_tag
        self._client = docker.from_env()
        self._container = None

    def _start_instance(self) -> None:
        self._container = self._client.containers.run(
            image=f"{self.docker_image.value}:{self.docker_image_tag}",
            command=f"{MEMGRAPH_DEFAULT_BINARY_PATH} {(' ').join([f'{k}={v}' for k, v in self.config.items()])}",
            detach=True,
            ports={f"{self.port}/tcp": self.port},
        )
        wait_for_docker_container(self._container)

    def _stop_instance(self) -> Dict:
        self._container.stop()

        return self._container.wait()

    def is_running(self) -> bool:
        """Check if the Memgraph instance is still running."""
        if self._container is None:
            return False

        self._container.reload()

        return self._container.status == DockerContainerStatus.RUNNING.value
